# Copyright 2024 Remy Blank <remy@c-space.org>
# SPDX-License-Identifier: MIT

import argparse
import contextlib
import datetime
from http import HTTPMethod, HTTPStatus
import itertools
import mimetypes
import os
import pathlib
import posixpath
import re
import shutil
import socket
import socketserver
import stat
import subprocess
import sys
import tempfile
import threading
import time
from urllib import parse, request
import webbrowser
from wsgiref import simple_server, util as wsgiutil

from . import __project__, __version__, api, deps, store, util, wsgi

# TODO: Split groups of sub-commands into separate modules

dev_store = pathlib.Path('tmp/store.sqlite')


@util.main
def main(argv, stdin, stdout, stderr):
    """Run the command."""
    parser = util.get_arg_parser(stdin, stdout, stderr)(
        prog=pathlib.Path(argv[0]).name, description="Manage a t-doc book.")
    root = parser.add_subparsers(title='Sub-commands')
    root.required = True

    p = root.add_parser('build', help="Build a book.")
    p.set_defaults(handler=cmd_build)
    arg = p.add_argument
    arg('target', metavar='TARGET', nargs='+', help="The build targets to run.")
    add_options(p, sphinx=True)

    p = root.add_parser('clean', help="Clean the build products of a book.")
    p.set_defaults(handler=cmd_clean)
    add_options(p, sphinx=True)

    add_group_commands(root)

    p = root.add_parser('serve', help="Serve a book locally.")
    p.set_defaults(handler=cmd_serve)
    arg = p.add_argument
    arg('--bind', metavar='ADDRESS', dest='bind', default='localhost',
        help="The address to bind the server to (default: %(default)s). "
             "Specify ALL to bind to all interfaces.")
    arg('--cache', metavar='PATH', type='path', dest='cache', default='_cache',
        help="The path to the cache directory (default: %(default)s).")
    arg('--delay', metavar='DURATION', type=float, dest='delay', default=1.0,
        help="The delay in seconds between detecting a source change and "
             "triggering a build (default: %(default)s).")
    arg('--exit-on-failure', action='store_true', dest='exit_on_failure',
        help="Terminate the server on build failure.")
    arg('--exit-on-idle', metavar='DURATION', type=float, dest='exit_on_idle',
        default=0.0,
        help="The time in seconds after the last connection closes when the "
             "server terminates (default: %(default)s).")
    arg('--ignore', metavar='REGEXP', type='regexp', dest='ignore',
        default=f'(^|{re.escape(os.sep)})__pycache__$',
        help="A regexp matching files and directories to ignore from watching "
             "(default: %(default)s).")
    arg('--full-builds', action='store_true', dest='full_builds',
        help="Perform full builds on source changes.")
    arg('--interval', metavar='DURATION', type=float, dest='interval',
        default=1.0,
        help="The interval in seconds at which to check for source changes "
             "(default: %(default)s).")
    arg('--open', action='store_true', dest='open',
        help="Open the site in a browser tab after the first build completes.")
    arg('--port', metavar='PORT', type=int, dest='port', default=8000,
        help="The port to bind the server to (default: %(default)s).")
    arg('--restart-on-change', action='store_true', dest='restart_on_change',
        help="Restart the server on changes.")
    add_store_option(arg)
    arg('--watch', metavar='PATH', type='path', action='append', dest='watch',
        default=[],
        help="Additional directories to watch for changes.")
    add_options(p, sphinx=True)

    add_store_commands(root)
    add_token_commands(root)
    add_user_commands(root)

    p = root.add_parser('version', help="Display version information.")
    p.set_defaults(handler=cmd_version)
    add_options(p)

    cfg = parser.parse_args(argv[1:])
    return cfg.handler(cfg)


def add_options(parser, sphinx=False):
    if sphinx: add_sphinx_options(parser)
    arg = parser.add_argument_group("Common options").add_argument
    arg('--color', dest='color', choices=['auto', 'false', 'true'],
        default='auto',
        help="Control the use of colors in output (default: %(default)s).")
    arg('--debug', action='store_true', dest='debug',
        help="Enable debug functionality.")


def add_sphinx_options(parser):
    arg = parser.add_argument_group("Sphinx options").add_argument
    arg('--build', metavar='PATH', type='path', dest='build', default='_build',
        help="The path to the build directory (default: %(default)s).")
    arg('--source', metavar='PATH', type='path', dest='source', default='docs',
        help="The path to the source files (default: %(default)s).")
    arg('--sphinx-opt', metavar='OPT', action='append', dest='sphinx_opts',
        default=[], help="Additional options to pass to sphinx-build.")


def add_store_option(arg):
    arg('--store', metavar='PATH', type='path', dest='store',
        default=os.environ.get('TDOC_STORE'),
        help="The path to the store database.")


@contextlib.contextmanager
def get_store(cfg, allow_mem=False, check_latest=True):
    if cfg.store is None and (ds := dev_store.resolve()).exists():
        cfg.store = ds
    if not allow_mem and not cfg.store:
        raise Exception("--store: No path specified")
    with store.Store(cfg.store) as st:
        if check_latest:
            with contextlib.closing(st.connect()) as db:
                version, latest = st.version(db)
                if version != latest:
                    o = cfg.stdout
                    o.write(f"""\
{o.LYELLOW}The store database must be upgraded:{o.NORM} version\
 {o.CYAN}{version}{o.NORM} => {o.CYAN}{latest}{o.NORM}
Would you like to perform the upgrade (y/n)? """)
                    o.flush()
                    resp = input().lower()
                    o.write("\n")
                    if resp not in ('y', 'yes', 'o', 'oui', 'j', 'ja'):
                        raise Exception("Store version mismatch "
                                        f"(current: {version}, want: {latest})")
                    upgrade_store(cfg, st, db, version, latest)
        yield st


@contextlib.contextmanager
def get_db(cfg):
    with get_store(cfg) as store, contextlib.closing(store.connect()) as db:
        yield db


def store_backup_path(cfg):
    suffix = datetime.datetime.now() \
                .isoformat('.', 'seconds').replace(':', '-')
    return cfg.store.with_name(f'{cfg.store.name}.{suffix}')


def upgrade_store(cfg, store, db, version, to_version):
    o = cfg.stdout
    backup = store_backup_path(cfg)
    o.write(f"Backing up store to {backup}\n")
    store.backup(db, backup)
    def on_version(v): o.write(f"Upgrading store to version {v}\n")
    store.upgrade(db, version=to_version, on_version=on_version)
    o.write("Store upgraded successfully\n")


def comma_separated(s):
    if s is None: return []
    return s.split(',')


def cmd_build(cfg):
    for target in cfg.target:
        res = sphinx_build(cfg, target, build=cfg.build)
        if res.returncode != 0: return res.returncode


def cmd_clean(cfg):
    return sphinx_build(cfg, 'clean', build=cfg.build).returncode


def add_group_commands(parser):
    p = parser.add_parser('group', help="Group-related commands.")
    sp = p.add_subparsers(title="Sub-commands")
    sp.required = True

    def add_groups(arg):
        for name in ('group', 'groups'):
            arg(f'--{name}', metavar="GROUP,...", dest='groups',
                help="A comma-separated list of groups.")

    def add_users(arg):
        for name in ('user', 'users'):
            arg(f'--{name}', metavar="USER,...", dest='users',
                help="A comma-separated list of users.")

    def add_groups_re(arg):
        arg('groups', metavar='REGEXP', nargs='?', default='',
            help="A regexp to limit the groups to consider.")

    p = sp.add_parser('add', help="Add members to one or more groups.")
    p.set_defaults(handler=cmd_group_add)
    arg = p.add_argument
    add_groups(arg)
    add_origin_option(arg)
    add_store_option(arg)
    add_users(arg)
    arg('group', metavar='GROUP', nargs='+', help="The group to add to.")
    add_options(p)

    p = sp.add_parser('list', help="List groups.")
    p.set_defaults(handler=cmd_group_list)
    arg = p.add_argument
    add_store_option(arg)
    add_groups_re(arg)
    add_options(p)

    p = sp.add_parser('members', help="List members of a group.")
    p.set_defaults(handler=cmd_group_members)
    arg = p.add_argument
    add_origin_option(arg)
    add_store_option(arg)
    arg('--transitive', action='store_true', dest='transitive',
        help="Include transitive memberships.")
    add_groups_re(arg)
    add_options(p)

    p = sp.add_parser('memberships', help="List group memberships for a group.")
    p.set_defaults(handler=cmd_group_memberships)
    arg = p.add_argument
    add_origin_option(arg)
    add_store_option(arg)
    add_groups_re(arg)
    add_options(p)

    p = sp.add_parser('remove', help="Remove members from one or more groups.")
    p.set_defaults(handler=cmd_group_remove)
    arg = p.add_argument
    add_groups(arg)
    add_origin_option(arg)
    add_store_option(arg)
    add_users(arg)
    arg('group', metavar='GROUP', nargs='+', help="The group to remove from.")
    add_options(p)


def cmd_group_add(cfg):
    with get_db(cfg) as db, db:
        db.groups.modify(cfg.origin, cfg.group,
                         add_users=comma_separated(cfg.users),
                         add_groups=comma_separated(cfg.groups))


def cmd_group_list(cfg):
    with get_db(cfg) as db, db:
        groups = db.groups.list(cfg.groups)
    groups.sort()
    o = cfg.stdout
    for group in groups: cfg.stdout.write(f"{o.CYAN}{group}{o.NORM}\n")


def cmd_group_members(cfg):
    with get_db(cfg) as db, db:
        members = db.groups.members(cfg.origin, cfg.groups,
                                    transitive=cfg.transitive)
    members.sort()
    wgroup = max((len(m[0]) for m in members), default=0)
    wname = max((len(m[2]) for m in members), default=0)
    o = cfg.stdout
    prev = None
    for group, typ, name, transitive in members:
        prefix = f"{o.CYAN}{group:{wgroup}}{o.NORM}" if group != prev \
                 else f"{'':{wgroup}}"
        if transitive:
            cfg.stdout.write(
                f"{prefix}  {typ:5} {o.LWHITE}{name:{wname}}{o.NORM}  "
                "(transitive)\n")
        else:
            cfg.stdout.write(f"{prefix}  {typ:5} {o.LWHITE}{name}{o.NORM}\n")
        prev = group


def cmd_group_memberships(cfg):
    with get_db(cfg) as db, db:
        memberships = db.groups.memberships(cfg.origin, cfg.groups)
    memberships.sort()
    wmember = max((len(m[0]) for m in memberships), default=0)
    o = cfg.stdout
    prev = None
    for member, group in memberships:
        prefix = f"{o.CYAN}{member:{wmember}}{o.NORM}" if member != prev \
                 else f"{'':{wmember}}"
        cfg.stdout.write(f"{prefix}  {o.LWHITE}{group}{o.NORM}\n")
        prev = member


def cmd_group_remove(cfg):
    with get_db(cfg) as db, db:
        db.groups.modify(cfg.origin, cfg.group,
                         remove_users=comma_separated(cfg.users),
                         remove_groups=comma_separated(cfg.groups))


def cmd_serve(cfg):
    for family, _, _, _, addr in socket.getaddrinfo(
            cfg.bind if cfg.bind != 'ALL' else None, cfg.port,
            type=socket.SOCK_STREAM, flags=socket.AI_PASSIVE):
        break

    class Server(ServerBase):
        address_family = family

    with Server(addr, RequestHandler) as srv, \
            get_store(cfg, allow_mem=True) as store, \
            api.Api(store, stderr=cfg.stderr) as api_, \
            Application(cfg, addr, srv, api_) as app:
        srv.set_app(app)
        try:
            srv.serve_forever()
        except KeyboardInterrupt:
            cfg.restart_on_change = False
            cfg.stderr.write("Interrupted, exiting\n")

    if cfg.restart_on_change:
        cfg.stdout.flush()
        cfg.stderr.flush()
        os.execv(sys.argv[0], sys.argv)
    return app.returncode


def add_store_commands(parser):
    p = parser.add_parser('store', help="Store-related commands.")
    sp = p.add_subparsers(title="Sub-commands")
    sp.required = True

    p = sp.add_parser('backup', help="Backup the store database.")
    p.set_defaults(handler=cmd_store_backup)
    arg = p.add_argument
    add_store_option(arg)
    arg('destination', metavar='PATH', type='path', nargs='?', default=None,
        help="The path to the backup copy. Defaults to the source database "
             "file with a date + time suffix.")
    add_options(p)

    p = sp.add_parser('create', help="Create the store database.")
    p.set_defaults(handler=cmd_store_create)
    arg = p.add_argument
    arg('--dev', action='store_true', dest='dev',
        help="Create the store for dev mode.")
    add_store_option(arg)
    arg('--version', metavar='VERSION', type=int, dest='version', default=None,
        help="The version at which to create the store (default: latest).")
    add_options(p)

    p = sp.add_parser('upgrade', help="Upgrade the store database.")
    p.set_defaults(handler=cmd_store_upgrade)
    arg = p.add_argument
    add_store_option(arg)
    arg('--version', metavar='VERSION', type=int, dest='version', default=None,
        help="The version to which to upgrade the store (default: latest).")
    add_options(p)


def cmd_store_backup(cfg):
    with get_store(cfg, check_latest=False) as store, \
            contextlib.closing(store.connect(params='mode=ro')) as db:
        if cfg.destination is None: cfg.destination = store_backup_path(cfg)
        store.backup(db, cfg.destination)


def cmd_store_create(cfg):
    if cfg.store is None and cfg.dev: cfg.store = dev_store.resolve()
    if not cfg.store: raise Exception("--store: No path specified")
    st = store.Store(cfg.store)
    st.path.parent.mkdir(parents=True, exist_ok=True)
    version = st.create(version=cfg.version, dev=cfg.dev)
    cfg.stdout.write(f"Store created (version: {version})\n")


def cmd_store_upgrade(cfg):
    with get_store(cfg, check_latest=False) as store, \
            contextlib.closing(store.connect()) as db:
        version, latest = store.version(db)
        if version == latest:
            cfg.stdout.write(
                f"Store is already up-to-date (version: {version})\n")
            return
        upgrade_store(cfg, store, db, version,
                      cfg.version if cfg.version is not None else latest)


def add_token_commands(parser):
    p = parser.add_parser('token', help="Token-related commands.")
    sp = p.add_subparsers(title="Sub-commands")
    sp.required = True

    p = sp.add_parser('create', help="Create tokens.")
    p.set_defaults(handler=cmd_token_create)
    arg = p.add_argument
    arg('--expire', metavar='TIME', type='timestamp', dest='expire',
        help="The token expiry timestamp.")
    add_origin_option(arg)
    add_store_option(arg)
    arg('user', metavar='USER', nargs='+',
        help="The users for whom to create tokens.")
    add_options(p)

    p = sp.add_parser('expire', help="Expire tokens.")
    p.set_defaults(handler=cmd_token_expire)
    arg = p.add_argument
    arg('--time', metavar='TIME', type='timestamp', dest='time',
        help="The time when the tokens should expire (default: now).")
    add_store_option(arg)
    arg('token', metavar='TOKEN', nargs='+',
        help="The users and / or tokens to expire.")
    add_options(p)

    p = sp.add_parser('list', help="List tokens.")
    p.set_defaults(handler=cmd_token_list)
    arg = p.add_argument
    arg('--expired', action='store_true', dest='expired',
        help="Include expired tokens.")
    add_origin_option(arg)
    add_store_option(arg)
    arg('users', metavar='REGEXP', nargs='?', default='',
        help="A regexp to limit the users to consider.")
    add_options(p)


def add_origin_option(arg):
    arg('--origin', metavar='URL', dest='origin', default='',
        help="The origin on which to operate.")


def cmd_token_create(cfg):
    with get_db(cfg) as db, db:
        tokens = db.tokens.create(cfg.user, cfg.expire)
    width = max((len(u) for u in cfg.user), default=0)
    o = cfg.stdout
    for user, token in zip(cfg.user, tokens):
        cfg.stdout.write(
            f"{o.CYAN}{user:{width}}{o.NORM} "
            f"{o.LBLUE}{cfg.origin}#?login={token}{o.NORM}\n")


def cmd_token_expire(cfg):
    with get_db(cfg) as db, db:
        db.tokens.expire(cfg.token, cfg.time)


def cmd_token_list(cfg):
    with get_db(cfg) as db, db:
        tokens = db.tokens.list(cfg.users, expired=cfg.expired)
    epoch = datetime.datetime.fromtimestamp(0)
    tokens.sort(key=lambda r: (r[0], r[3], r[4] or epoch, r[2]))
    wuser = max((len(u) for u, *_ in tokens), default=0)
    o = cfg.stdout
    for user, uid, token, created, expires in tokens:
        if expires: expires = f", expires: {expires.isoformat(' ', 'seconds')}"
        cfg.stdout.write(
            f"{o.CYAN}{user:{wuser}}{o.NORM} "
            f"{o.LBLUE}{cfg.origin}#?login={token}{o.NORM}\n"
            f"  created: {created.isoformat(' ', 'seconds')}{expires or ""}\n")


def add_user_commands(parser):
    p = parser.add_parser('user', help="User-related commands.")
    sp = p.add_subparsers(title="Sub-commands")
    sp.required = True

    def add_users_re(arg):
        arg('users', metavar='REGEXP', nargs='?', default='',
            help="A regexp to limit the users to consider.")

    p = sp.add_parser('create', help="Create users.")
    p.set_defaults(handler=cmd_user_create)
    arg = p.add_argument
    add_origin_option(arg)
    add_store_option(arg)
    arg('--token-expire', metavar='TIME', type='timestamp', dest='token_expire',
        help="The expiry time of the users' token.")
    arg('user', metavar='USER', nargs='+',
        help="The names of the users to create.")
    add_options(p)

    p = sp.add_parser('list', help="List users.")
    p.set_defaults(handler=cmd_user_list)
    arg = p.add_argument
    add_store_option(arg)
    add_users_re(arg)
    add_options(p)

    p = sp.add_parser('memberships', help="List group memberships for a user.")
    p.set_defaults(handler=cmd_user_memberships)
    arg = p.add_argument
    add_origin_option(arg)
    add_store_option(arg)
    arg('--transitive', action='store_true', dest='transitive',
        help="Include transitive memberships.")
    add_users_re(arg)
    add_options(p)


def cmd_user_create(cfg):
    with get_db(cfg) as db, db:
        uids = db.users.create(cfg.user)
        tokens = db.tokens.create(cfg.user, cfg.token_expire)
    wuser = max((len(u) for u in cfg.user), default=0)
    o = cfg.stdout
    for user, uid, token in zip(cfg.user, uids, tokens):
        cfg.stdout.write(f"{o.CYAN}{user:{wuser}}{o.NORM} ({uid:016x})  "
                         f"{o.LBLUE}{cfg.origin}#?login={token}{o.NORM}\n")


def cmd_user_list(cfg):
    with get_db(cfg) as db, db:
        users = db.users.list(cfg.users)
    users.sort(key=lambda r: r[0])
    wuser = max((len(u[0]) for u in users), default=0)
    o = cfg.stdout
    for user, uid, created in users:
        cfg.stdout.write(
            f"{o.CYAN}{user:{wuser}}{o.NORM} ({uid:016x})  "
            f"created: {created.isoformat(' ', 'seconds')}\n")


def cmd_user_memberships(cfg):
    with get_db(cfg) as db, db:
        memberships = db.users.memberships(cfg.origin, cfg.users,
                                           transitive=cfg.transitive)
    memberships.sort()
    wuser = max((len(m[0]) for m in memberships), default=0)
    wgroup = max((len(m[1]) for m in memberships), default=0)
    o = cfg.stdout
    prev = None
    for user, group, transitive in memberships:
        prefix = f"{o.CYAN}{user:{wuser}}{o.NORM}" if user != prev \
                 else f"{'':{wuser}}"
        if transitive:
            cfg.stdout.write(f"{prefix}  {o.LWHITE}{group:{wgroup}}{o.NORM}  "
                             "(transitive)\n")
        else:
            cfg.stdout.write(f"{prefix}  {o.LWHITE}{group}{o.NORM}\n")
        prev = user


def cmd_version(cfg):
    cfg.stdout.write(f"{__project__}-{__version__}\n")


def sphinx_build(cfg, target, *, build, tags=(), **kwargs):
    argv = [sys.executable, '-P', '-m', 'sphinx', 'build', '-M', target,
            cfg.source, build, '--fail-on-warning', '--jobs=auto']
    argv += [f'--tag={tag}' for tag in tags]
    if cfg.debug: argv += ['--show-traceback']
    argv += cfg.sphinx_opts
    return subprocess.run(argv, stdin=cfg.stdin, stdout=cfg.stdout,
                          stderr=cfg.stderr, **kwargs)


class ServerBase(socketserver.ThreadingMixIn, simple_server.WSGIServer):
    daemon_threads = True

    def server_bind(self):
        with contextlib.suppress(Exception):
            self.socket.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_V6ONLY, 0)
        return super().server_bind()


class RequestHandler(simple_server.WSGIRequestHandler):
    def log_request(self, code='-', size='-'):
        pass

    def log_message(self, format, *args):
        self.server.application.cfg.stderr.write("%s - - [%s] %s\n" % (
            self.address_string(), self.log_date_time_string(),
            (format % args).translate(self._control_char_table)))


def try_stat(path):
    with contextlib.suppress(OSError):
        return path.stat()


class Application:
    def __init__(self, cfg, addr, server, api_):
        self.cfg = cfg
        self.addr = addr
        self.server = server
        self.lock = threading.Condition(threading.Lock())
        self.directory = self.build_dir(0) / 'html'
        self.stop = False
        self.min_mtime = time.time_ns()
        self.returncode = 0
        self.api = api_
        self.api.add_endpoint('terminate', self.handle_terminate)
        self.opened = False
        self.build_mtime = None
        self.build_obs = api.ValueObservable('build', self.build_mtime)
        self.api.events.add_observable(self.build_obs)
        self.builder = threading.Thread(target=self.watch_and_build)
        self.builder.start()
        self.endpoints = {
            '_api': self.api,
            '_cache': self.handle_cache,
        }

    def __enter__(self): return self

    def __exit__(self, typ, value, tb):
        with self.lock:
            self.stop = True
            self.lock.notify()
        self.builder.join()

    def watch_and_build(self):
        self.remove_all()
        interval = self.cfg.interval * 1_000_000_000
        delay = self.cfg.delay * 1_000_000_000
        idle = self.cfg.exit_on_idle * 1_000_000_000
        prev, prev_mtime, build_mtime = 0, 0, None
        build_next = self.build_dir('next')
        # TODO: Use monotonic clock for delays
        # TODO: Avoid looping at 10Hz
        while True:
            with self.lock:
                if prev != 0: self.lock.wait_for(lambda: self.stop, timeout=0.1)
                if self.stop: break
            now = time.time_ns()
            if (idle > 0 and (lw := self.api.events.last_watcher) is not None
                    and now > lw + idle):
                self.server.shutdown()
                break
            if now < prev + interval: continue
            mtime = self.latest_mtime()
            if mtime <= prev_mtime:
                prev = now
                continue
            if now < mtime + delay and prev_mtime != 0:
                prev = mtime + delay - interval
                continue
            if prev_mtime != 0:
                if self.cfg.restart_on_change:
                    self.cfg.stdout.write(
                        "\nSource change detected, restarting\n")
                    self.server.shutdown()
                    break
                self.cfg.stdout.write(
                    "\nSource change detected, rebuilding\n")
            prev_mtime = mtime
            if self.build(build_next):
                build = self.build_dir(mtime)
                os.rename(build_next, build)
                with self.lock:
                    self.build_mtime = mtime
                    self.directory = build / 'html'
                self.build_obs.set(str(mtime))
                self.print_serving()
                if build_mtime is not None:
                    self.remove(self.build_dir(build_mtime))
                build_mtime = mtime
            else:
                self.remove(build_next)
            if not self.cfg.full_builds and build_mtime is not None:
                shutil.copytree(self.build_dir(build_mtime), build_next,
                                symlinks=True)
            self.print_upgrade()
            prev = time.time_ns()
        if build_mtime is not None: self.remove(self.build_dir(build_mtime))
        self.remove(build_next)

    def latest_mtime(self):
        def on_error(e):
            self.cfg.stderr.write(f"Scan: {e}\n")
        mtime = self.min_mtime
        for path in itertools.chain([self.cfg.source], self.cfg.watch):
            for base, dirs, files in path.walk(on_error=on_error):
                for file in files:
                    p = base / file
                    if self.cfg.ignore.search(str(p)) is not None: continue
                    try:
                        st = p.stat()
                        if stat.S_ISREG(st.st_mode):
                            mtime = max(mtime, st.st_mtime_ns)
                    except Exception as e:
                        on_error(e)
                dirs[:] = [d for d in dirs
                           if self.cfg.ignore.search(str(base / d)) is None]
        return mtime

    def build_dir(self, mtime):
        return self.cfg.build / f'serve-{mtime}'

    def build(self, build):
        try:
            res = sphinx_build(self.cfg, 'html', build=build,
                               tags=['tdoc-dev'])
            if res.returncode == 0: return True
        except Exception as e:
            self.cfg.stderr.write(f"Build: {e}\n")
        if self.cfg.exit_on_failure:
            self.returncode = 1
            self.server.shutdown()
        return False

    def remove(self, build):
        build.relative_to(self.cfg.build)  # Ensure we're below the build dir
        if not build.exists(): return
        def on_error(fn, path, e):
            self.cfg.stderr.write(f"Removal: {fn}: {path}: {e}\n")
        shutil.rmtree(build, onexc=on_error)

    def remove_all(self):
        for build in self.cfg.build.glob('serve-*'):
            self.remove(build)

    def print_serving(self):
        host, port = self.addr[:2]
        if ':' in host: host = f'[{host}]'
        o = self.cfg.stdout
        o.write(f"Serving at <{o.LBLUE}http://{host}:{port}/{o.NORM}>\n")
        o.flush()
        if self.cfg.open and not self.opened:
            self.opened = True
            webbrowser.open_new_tab(f'http://{host}:{port}/')

    def print_upgrade(self):
        if sys.prefix == sys.base_prefix: return  # Not running in a venv
        o = self.cfg.stdout
        with contextlib.suppress(Exception):
            marker = pathlib.Path(sys.prefix) / 'upgrade.txt'
            cur, new = marker.read_text('utf-8').split(' ')[:2]
            if cur == new: return
            o.write(f"""\
{o.LYELLOW}An upgrade is available:{o.NORM} {__project__}\
 {o.CYAN}{cur}{o.NORM} => {o.CYAN}{new}{o.NORM}
Release notes: <{o.LBLUE}https://common.t-doc.org/release-notes.html\
#release-{new.replace('.', '-')}{o.NORM}>
{o.LWHITE}Restart the server to upgrade.{o.NORM}
""")

    def __call__(self, env, respond):
        try:
            env['tdoc.dev'] = True

            # Dispatch endpoints.
            script_name, path_info = env['SCRIPT_NAME'], env['PATH_INFO']
            part = wsgiutil.shift_path_info(env)
            if (handler := self.endpoints.get(part)) is not None:
                yield from handler(env, respond)
                return
            env['SCRIPT_NAME'], env['PATH_INFO'] = script_name, path_info

            # Serve from the filesystem.
            with self.lock: base = self.directory
            yield from self.handle_file(env, respond, base)
        except wsgi.Error as e:
            yield from wsgi.error(respond, e.status, e.message,
                                  exc_info=sys.exc_info())

    def handle_cache(self, env, respond):
        yield from self.handle_file(env, respond, self.cfg.cache,
                                    self.on_cache_not_found)

    def on_cache_not_found(self, path_info, path):
        try:
            parts = path_info.split('/', 3)
            if parts[0] != '' or len(parts) < 4: return
            if (d := deps.info.get(parts[1])) is None: return
            url = f'{d['url'](parts[2])}/{parts[3]}'
            self.cfg.stderr.write(f"Caching {url}\n")
            with request.urlopen(url) as f: data = f.read()
            path.parent.mkdir(parents=True, exist_ok=True)
            with tempfile.NamedTemporaryFile(
                    dir=path.parent, prefix=path.name + '.',
                    delete_on_close=False) as f:
                f.write(data)
                f.close()
                pathlib.Path(f.name).replace(path)
        except Exception as e:
            self.cfg.stderr.write(f"Cache: {e}\n")

    def handle_file(self, env, respond, base, on_not_found=None):
        env['wsgi.multithread'] = True
        method = wsgi.method(env, HTTPMethod.HEAD, HTTPMethod.GET)
        path_info = env['PATH_INFO']
        path = self.file_path(path_info, base)
        path.relative_to(base)  # Ensure we're below base
        if (st := try_stat(path)) is None:
            if on_not_found is None: raise wsgi.Error(HTTPStatus.NOT_FOUND)
            on_not_found(path_info, path)
            if (st := try_stat(path)) is None:
                raise wsgi.Error(HTTPStatus.NOT_FOUND)

        if stat.S_ISDIR(st.st_mode):
            parts = parse.urlsplit(path_info)
            if not parts.path.endswith('/'):
                location = parse.urlunsplit(
                    (parts[:2] + (parts[2] + '/',) + parts[3:]))
                respond(wsgi.http_status(HTTPStatus.MOVED_PERMANENTLY), [
                    ('Location', location),
                    ('Content-Length', '0'),
                ])
                return
            path = path / 'index.html'
            if (st := try_stat(path)) is None:
                raise wsgi.Error(HTTPStatus.NOT_FOUND)

        if not stat.S_ISREG(st.st_mode):
            raise wsgi.Error(HTTPStatus.NOT_FOUND)
        mime_type = mimetypes.guess_type(path)[0] or 'application/octet-stream'
        respond(wsgi.http_status(HTTPStatus.OK), [
            ('Content-Type', mime_type),
            ('Content-Length', str(st.st_size)),
        ])
        if method == HTTPMethod.HEAD: return
        wrapper = env.get('wsgi.file_wrapper', wsgiutil.FileWrapper)
        yield from wrapper(open(path, 'rb'))

    def file_path(self, path, base):
        trailing = path.rstrip().endswith('/')
        try:
            path = parse.unquote(path, errors='surrogatepass')
        except UnicodeDecodeError:
            path = parse.unquote(path)
        res = base
        for part in filter(None, posixpath.normpath(path).split('/')):
            if pathlib.Path(part).parent.name or part in (os.curdir, os.pardir):
                continue
            res = res / part
        return res / '' if trailing else res

    def handle_terminate(self, env, respond):
        wsgi.method(env, HTTPMethod.POST)
        if env['PATH_INFO']: raise wsgi.Error(HTTPStatus.NOT_FOUND)
        req = wsgi.read_json(env)
        rc = req.get('rc', 0)
        yield from wsgi.respond_json(respond, {})
        try:
            self.returncode = int(rc)
        except ValueError:
            self.returncode = 1
        self.server.shutdown()


if __name__ == '__main__':
    main()



# Copyright 2024 Remy Blank <remy@c-space.org>
# SPDX-License-Identifier: MIT

import sys

from tdoc.common.defaults import *

project = "t-doc"
copyright = "%Y Remy Blank"
license = 'MIT'
language = 'en'

# myst_links_external_new_tab = True

html_theme_options = {
    'repository_url': 'https://github.com/t-doc-org/common',
    'use_download_button': True,
}

tdoc_enable_sab = 'no'
# tdoc_enable_sab = 'cross-origin-isolation'
# tdoc_enable_sab = 'sabayon'

intersphinx_mapping = {
    'python': (
        f'https://docs.python.org/{sys.version_info[0]}.{sys.version_info[1]}',
        None,
    ),
    'sphinx': ('https://www.sphinx-doc.org/en/master', None),
}



# Copyright 2024 Remy Blank <remy@c-space.org>
# SPDX-License-Identifier: MIT

from docutils import languages, nodes
from docutils.parsers.rst import directives
from docutils.parsers.rst.directives import admonitions
from sphinx import config
from sphinx.util import logging

from . import _, __version__

_log = logging.getLogger(__name__)


def setup(app):
    app.add_config_value('tdoc_solutions', 'show', 'html',
                         config.ENUM('show', 'hide', 'dynamic'))
    app.add_directive('solution', Solution)
    app.add_node(solution, html=(visit_solution, depart_solution))
    app.connect('tdoc-html-page-config', set_html_page_config)
    app.connect('html-page-context', add_header_button, priority=500.5)
    return {
        'version': __version__,
        'parallel_read_safe': True,
        'parallel_write_safe': True,
    }


class solution(nodes.Admonition, nodes.Element): pass


class Solution(admonitions.BaseAdmonition):
    node_class = solution
    optional_arguments = 1
    option_spec = admonitions.BaseAdmonition.option_spec | {
        'expand': directives.flag,
        'show': directives.flag,
    }

    def run(self):
        res = super().run()
        title_text = self.arguments[0] if self.arguments else _("Solution")
        for node in res:
            if not isinstance(node, self.node_class): continue
            text_nodes, msgs = self.state.inline_text(title_text, self.lineno)
            title = nodes.title(title_text, '', *text_nodes)
            title.source, title.line = \
                self.state_machine.get_source_and_line(self.lineno)
            node[:0] = [title] + msgs
            cls = node['classes']
            if not cls: cls += ['note', 'dropdown']
            cls += ['solution']
            if 'expand' in self.options: cls += ['expand']
            if 'show' in self.options: cls += ['always-show']
        return res


def set_html_page_config(app, page, config, doctree):
    if page is None: return
    md = app.env.metadata[page]
    v = md.get('solutions', app.config.tdoc_solutions)
    if not app.config.values['tdoc_solutions'].valid_types.match(v):
        _log.warning(f"{{metadata}}: Invalid 'solutions' value: {v}")
        v = md['solutions'] = 'show'
    if v != 'show':
        config['html_data']['tdocSolutions'] = v


def add_header_button(app, page, template, context, doctree):
    if doctree is None: return
    if all('always-show' in sol['classes']
           for sol in doctree.findall(solution)): return
    context["header_buttons"].append({
        'type': 'javascript',
        'javascript': 'tdoc.toggleSolutions()',
        'tooltip': _("Toggle solutions"),
        'label': 'toggle-solutions',
    })


def visit_solution(self, node):
    return self.visit_admonition(node)


def depart_solution(self, node):
    return self.depart_admonition(node)



# Copyright 2024 Remy Blank <remy@c-space.org>
# SPDX-License-Identifier: MIT

import contextlib
import hashlib
from http import HTTPMethod, HTTPStatus
import json
import queue
import secrets
import sys
import threading
import time
import traceback
from wsgiref import util

from . import store, wsgi

missing = object()
# TODO(py-3.13): Remove Shutdown
Shutdown = getattr(queue, 'Shutdown', queue.Empty)


def arg(data, name, validate=None):
    if (v := data.get(name, missing)) is missing:
        raise wsgi.Error(HTTPStatus.BAD_REQUEST, f"Missing argument '{name}'")
    if validate is not None and not validate(v):
        raise wsgi.Error(HTTPStatus.BAD_REQUEST, f"Invalid argument '{name}'")
    return v


def args(data, *names):
    return tuple(arg(data, n) for n in names)


def check(cond, code=HTTPStatus.FORBIDDEN, msg=None):
    if not cond: raise wsgi.Error(code, msg)


class Api:
    def __init__(self, store, stderr=None, db_pool_size=16):
        if stderr is None: stderr = sys.stderr
        self.stderr = stderr
        self.store = store
        self.pool = store.pool(size=db_pool_size)
        self.events = EventsApi(self)
        self.endpoints = {
            'events': self.events,
            'health': self.handle_health,
            'log': self.handle_log,
            'poll': self.handle_poll,
            'solutions': self.handle_solutions,
            'user': self.handle_user,
        }

    def stop(self):
        self.events.stop()

    def __enter__(self): return self

    def __exit__(self, typ, value, tb):
        self.stop()

    def add_endpoint(self, name, handler):
        self.endpoints[name] = handler

    def print_exception(self, e=None, limit=None, chain=True):
        if e is None: e = sys.exception()
        traceback.print_exception(e, limit=limit, chain=chain, file=self.stderr)

    def db(self, env):
        if (db := env.get('tdoc.db')) is not None: return db
        db = env['tdoc.db'] = self.pool.get()
        return db

    def authenticate(self, env):
        user = None
        if token := wsgi.authorization(env):
            with self.db(env) as db:
                user = db.tokens.authenticate(token)
            if user is None: raise wsgi.Error(HTTPStatus.UNAUTHORIZED)
        env['tdoc.user'] = user

    def user(self, env, anon=True):
        user = env.get('tdoc.user')
        if user is None and not anon: raise wsgi.Error(HTTPStatus.UNAUTHORIZED)
        return user

    def member_of(self, env, db, group):
        return db.users.member_of(wsgi.origin(env), self.user(env), group)

    def check_acl(self, env, perm):
        token = wsgi.authorization(env)
        with self.db(env) as db:
            for perms, in db.execute(
                    "select perms from auth where token in (?, '*')", (token,)):
                perms = perms.split(',')
                if perm in perms or '*' in perms: return
        raise wsgi.Error(HTTPStatus.FORBIDDEN)

    def __call__(self, env, respond):
        try:
            name = util.shift_path_info(env)
            if (handler := self.endpoints.get(name)) is None:
                raise wsgi.Error(HTTPStatus.NOT_FOUND)
            self.authenticate(env)
            yield from handler(env, respond)
        except wsgi.Error as e:
            yield from wsgi.error(respond, e.status, e.message,
                                  exc_info=sys.exc_info())
        except store.client_errors:
            self.print_exception(limit=-1, chain=False)
            yield from wsgi.error(respond, HTTPStatus.BAD_REQUEST,
                                  exc_info=sys.exc_info())
        except store.Error as e:
            self.print_exception(limit=-1, chain=False)
            msg = e.args[0] if e.args else None
            yield from wsgi.error(respond, HTTPStatus.FORBIDDEN,
                                  msg, exc_info=sys.exc_info())
        finally:
            if (db := env.get('tdoc.db')) is not None: self.pool.release(db)

    def handle_health(self, env, respond):
        wsgi.method(env, HTTPMethod.GET)
        if env['PATH_INFO']: raise wsgi.Error(HTTPStatus.NOT_FOUND)
        return wsgi.respond_json(respond, {})

    def handle_log(self, env, respond):
        wsgi.method(env, HTTPMethod.POST)
        if env['PATH_INFO']: raise wsgi.Error(HTTPStatus.NOT_FOUND)
        self.check_acl(env, 'log')
        req = wsgi.read_json(env)
        with self.db(env) as db:
            db.execute("""
                insert into log (time, location, session, data)
                    values (?, ?, ?, json(?))
            """, (int(req.get('time', time.time_ns() // 1_000_000)),
                  req['location'], req.get('session'),
                  wsgi.to_json(req['data'])))
        return wsgi.respond_json(respond, {})

    def handle_poll(self, env, respond):
        wsgi.method(env, HTTPMethod.POST)
        if env['PATH_INFO']: raise wsgi.Error(HTTPStatus.NOT_FOUND)
        origin = wsgi.origin(env)
        req = wsgi.read_json(env)
        with self.db(env) as db:
            if polls := req.get('open'):
                check(self.member_of(env, db, 'polls:control'))
                for poll in polls:
                    mode = arg(poll, 'mode', lambda v: v in ('single', 'multi'))
                    expires = None if (exp := poll.get('exp')) is None \
                              else time.time_ns() + exp * 1_000_000
                    db.polls.open(origin, arg(poll, 'id'), mode,
                                  arg(poll, 'answers'), expires)
            if ids := req.get('close'):
                check(self.member_of(env, db, 'polls:control'))
                db.polls.close(origin, ids)
            if ids := req.get('results'):
                check(self.member_of(env, db, 'polls:control'))
                db.polls.results(origin, ids, arg(req, 'value'))
            if ids := req.get('solutions'):
                check(self.member_of(env, db, 'polls:control'))
                db.polls.solutions(origin, ids, arg(req, 'value'))
            if ids := req.get('clear'):
                check(self.member_of(env, db, 'polls:control'))
                db.polls.clear(origin, ids)
            if 'vote' in req:
                db.polls.vote(origin, *args(req, 'id', 'voter', 'answer',
                                            'vote'))
        return wsgi.respond_json(respond, {})

    def handle_solutions(self, env, respond):
        wsgi.method(env, HTTPMethod.POST)
        if env['PATH_INFO']: raise wsgi.Error(HTTPStatus.NOT_FOUND)
        origin = wsgi.origin(env)
        req = wsgi.read_json(env)
        page = arg(req, 'page')
        show = arg(req, 'show', lambda v: v in ('show', 'hide'))
        with self.db(env) as db:
            check(self.member_of(env, db, 'solutions:write'))
            db.solutions.set_show(origin, page, show)
        return wsgi.respond_json(respond, {})

    def handle_user(self, env, respond):
        wsgi.method(env, HTTPMethod.POST)
        if env['PATH_INFO']: raise wsgi.Error(HTTPStatus.NOT_FOUND)
        origin = wsgi.origin(env)
        user = self.user(env, anon=False)
        with self.db(env) as db:
            info = db.users.info(origin, user)
        return wsgi.respond_json(respond, info)


class EventsApi:
    def __init__(self, api):
        self.api = api
        self.endpoints = {
            'sub': self.handle_sub,
            'watch': self.handle_watch,
        }
        self.lock = threading.Lock()
        self.observables = {}
        self.dyn_observables = {
            'solutions': SolutionsObservable,
            'poll': PollObservable,
            'poll/votes': PollVotesObservable,
        }
        self.watchers = {}
        self._stop_watchers = False
        self._last_watcher = None

    def stop(self):
        with self.lock:
            self._stop_watchers = True
            for w in self.watchers.values(): w.stop()

    def __call__(self, env, respond):
        name = util.shift_path_info(env)
        if (handler := self.endpoints.get(name)) is None:
            raise wsgi.Error(HTTPStatus.NOT_FOUND)
        return handler(env, respond)

    def add_observable(self, obs):
        with self.lock: self.observables[obs.key] = obs

    def remove_observable(self, obs):
        with self.lock: self.observables.pop(obs.key, None)

    def find_observable(self, req, env):
        key = Observable.hash(req)
        with self.lock:
            if (obs := self.observables.get(key)) is not None:
                if not obs.stopping: return obs
            if (typ := self.dyn_observables.get(req['name'])) is not None:
                obs = typ(req, self, env)
                self.observables[obs.key] = obs
                return obs
        raise Exception("Observable not found")

    def watchers_must_stop(self):
        with self.lock: return self._stop_watchers

    @property
    def last_watcher(self):
        with self.lock: return self._last_watcher

    @contextlib.contextmanager
    def watcher(self):
        with Watcher(self.watchers_must_stop) as watcher:
            sid = secrets.token_urlsafe(8)
            with self.lock:
                if self._stop_watchers:
                    raise wsgi.Error(HTTPStatus.REQUEST_TIMEOUT)
                while sid in self.watchers:
                    sid = secrets.token_urlsafe(8)
                watcher.sid = sid
                self.watchers[sid] = watcher
                self._last_watcher = None
            try:
                yield watcher
            finally:
                with self.lock:
                    del self.watchers[sid]
                    if not self.watchers:
                        self._last_watcher = time.time_ns()

    def handle_watch(self, env, respond):
        wsgi.method(env, HTTPMethod.POST)
        if env['PATH_INFO']: raise wsgi.Error(HTTPStatus.NOT_FOUND)
        req = wsgi.read_json(env)
        with self.watcher() as watcher:
            respond(wsgi.http_status(HTTPStatus.OK), [
                ('Content-Type', 'text/plain; charset=utf-8'),
                ('Cache-Control', 'no-store'),
            ])
            resp = {'sid': watcher.sid}
            if failed := self.watch(watcher, req.get('add', []), env):
                resp['failed'] = failed
            yield wsgi.to_json(resp).encode('utf-8') + b'\n'
            yield from watcher

    def handle_sub(self, env, respond):
        wsgi.method(env, HTTPMethod.POST)
        if env['PATH_INFO']: raise wsgi.Error(HTTPStatus.NOT_FOUND)
        req = wsgi.read_json(env)
        sid = arg(req, 'sid')
        with self.lock: watcher = self.watchers.get(sid)
        if watcher is None:
            raise wsgi.Error(HTTPStatus.BAD_REQUEST, "Unknown stream ID")
        resp = {}
        for wid in req.get('remove', []): watcher.unwatch(wid)
        if failed := self.watch(watcher, req.get('add', []), env):
            resp['failed'] = failed
        return wsgi.respond_json(respond, resp)

    def watch(self, watcher, adds, env):
        failed = []
        for add in adds:
            if (wid := add.get('wid')) is None: continue
            try:
                watcher.watch(wid, self.find_observable(add['req'], env))
            except Exception:
                failed.append(wid)
        return failed


class Watcher:
    def __init__(self, must_stop):
        self.must_stop = must_stop
        self.queue = queue.Queue()
        self.lock = threading.Lock()
        self.watches = {}

    def stop(self):
        # TODO(py-3.13): Remove must_stop and this conditional
        if hasattr(self.queue, 'shutdown'):
            self.queue.shutdown(True)

    def send(self, wid, msg):
        self.queue.put((wid, msg))

    def __iter__(self):
        while not self.must_stop():
            try:
                wid, msg = self.queue.get(timeout=1)
                yield b'{"wid":%d,"data":' % wid
                yield msg
                yield b'}\n'
            except queue.Empty:
                yield b'\n'
            except Shutdown:
                return

    def __enter__(self): return self

    def __exit__(self, typ, value, tb):
        with self.lock:
            watches, self.watches = self.watches, {}
        for wid, obs in watches.items(): obs.unwatch(self, wid)

    def watch(self, wid, obs):
        with self.lock:
            if wid in self.watches: return
            self.watches[wid] = obs
        obs.watch(self, wid)

    def unwatch(self, wid):
        with self.lock: obs = self.watches.pop(wid, None)
        if obs is not None: obs.unwatch(self, wid)


class Observable:
    @staticmethod
    def hash(req):
        return hashlib.sha256(
            wsgi.to_json(req, sort_keys=True).encode('utf-8')).digest()

    def __init__(self, req):
        self.key = self.hash(req)
        self.lock = threading.Condition(threading.Lock())
        self.watches = set()

    def send_initial_locked(self, watcher, wid): pass

    stopping = False

    def stop_locked(self): pass

    def send_locked(self, msg):
        for watcher, wid in self.watches: watcher.send(wid, msg)

    def watch(self, watcher, wid):
        key = (watcher, wid)
        with self.lock:
            if key in self.watches: return
            self.watches.add(key)
            self.send_initial_locked(watcher, wid)

    def unwatch(self, watcher, wid):
        with self.lock:
            self.watches.discard((watcher, wid))
            if not self.watches: self.stop_locked()


class ValueObservable(Observable):
    def __init__(self, name, value):
        super().__init__({'name': name})
        self._value = value

    def set(self, value):
        with self.lock:
            if value == self._value: return
            self._value = value
            self.send_locked(self._msg())

    def _msg(self):
        return wsgi.to_json(self._value).encode('utf-8')

    def send_initial_locked(self, watcher, wid):
        watcher.send(wid, self._msg())


class DynObservable(Observable):
    def __init__(self, req, events):
        super().__init__(req)
        self.events = events

    def unwatch(self, watcher, wid):
        super().unwatch(watcher, wid)
        if self.stopping: self.remove()

    def remove(self):
        self.events.remove_observable(self)

    def print_exception(self, e=None, limit=None, chain=True):
        self.events.api.print_exception(e=e, limit=limit, chain=chain)


def limit_interval(interval, burst=1):
    interval = int(interval * 1_000_000_000)
    count, prev = 0, time.monotonic_ns()

    def limit():
        nonlocal count, prev
        now = time.monotonic_ns()
        count = count - min(count, (now - prev) // interval) + 1
        if count < burst:
            prev = now
            return 0
        count = burst
        prev = now + interval
        return interval / 1_000_000_000

    return limit


class DbObservable(DynObservable):
    def __init__(self, req, events, data=None, limit=None):
        super().__init__(req, events)
        self._data = data
        self._limit = limit if limit is not None else limit_interval(1, burst=4)
        self._stop = False
        self._poller = threading.Thread(target=self.poll)
        self._poller.start()

    def _msg(self):
        return wsgi.to_json(self._data).encode('utf-8')

    def send_initial_locked(self, watcher, wid):
        if self._data is not None: watcher.send(wid, self._msg())

    @property
    def stopping(self):
        with self.lock: return self._stop

    def stop_locked(self):
        self._stop = True
        self.lock.notify()

    def wake_keys(self, db): return None

    def poll(self):
        try:
            store = self.events.api.store
            with contextlib.closing(store.connect(params='mode=ro')) as db, \
                    store.waker(self.lock, self.wake_keys(db), db,
                                self._limit) as waker:
                while True:
                    queried = False
                    try:
                        with db: data, until = self.query(db)
                        queried = True
                    except Exception:
                        self.print_exception()
                    with self.lock:
                        if queried and data != self._data:
                            self._data = data
                            self.send_locked(self._msg())
                        waker.wait(lambda: self._stop, until)
                        if self._stop: break
        except Exception:
            with self.lock: self._stop = True
            self.remove()
            self.print_exception()

    def query(self, db):
        raise NotImplementedError()


class SolutionsObservable(DbObservable):
    def __init__(self, req, events, env):
        self._origin = wsgi.origin(env)
        self._page = arg(req, 'page')
        super().__init__(req, events)

    def wake_keys(self, db):
        return [db.solutions.show_key(self._origin, self._page)]

    def query(self, db):
        return {'show': db.solutions.get_show(self._origin, self._page)}, None


class PollObservable(DbObservable):
    def __init__(self, req, events, env):
        self._origin = wsgi.origin(env)
        self._id = arg(req, 'id')
        super().__init__(req, events)

    def wake_keys(self, db):
        return [db.polls.poll_key(self._origin, self._id)]

    def query(self, db):
        data = db.polls.poll_data(self._origin, self._id)
        if (exp := data.pop('exp')) is not None and exp <= time.time_ns():
            exp = None
        return data, exp


class PollVotesObservable(DbObservable):
    def __init__(self, req, events, env):
        self._origin = wsgi.origin(env)
        self._voter, self._ids = args(req, 'voter', 'ids')
        self._ids.sort()
        super().__init__(req, events)

    def wake_keys(self, db):
        return [db.polls.voter_key(self._origin, poll, self._voter)
                for poll in self._ids]

    def query(self, db):
        return db.polls.votes_data(self._origin, self._voter, self._ids), None



# Copyright 2024 Remy Blank <remy@c-space.org>
# SPDX-License-Identifier: MIT

from docutils import nodes
from sphinx import errors
from sphinx.environment import collectors
from sphinx.util import docutils, logging, nodes as sphinx_nodes

from . import __version__

_log = logging.getLogger(__name__)


def setup(app):
    app.add_role('num', Num())
    app.add_enumerable_node(num, 'num', lambda n: '<num_title>',
                            html=(visit_num, depart_num))
    app.connect('config-inited', update_numfig_format)
    app.connect('builder-inited', NumCollector.init)
    app.add_env_collector(NumCollector)
    app.connect('env-get-updated', number_per_namespace, priority=501)
    app.connect('doctree-resolved', update_num_nodes)
    return {
        'version': __version__,
        'parallel_read_safe': True,
        'parallel_write_safe': True,
    }


class NoNum:
    def __contains__(self, value): return False
    def __mod__(self, other): return ''
    def __str__(self): return 'NoNum'
    def __eq__(self, other): return isinstance(other, NoNum)
    def __ne__(self, other): return not isinstance(other, NoNum)

no_num = NoNum()


def update_numfig_format(app, config):
    # Disable numbering by default for all standard enumerable node types except
    # 'section', and set the default format for 'num'.
    numfig_format = config.numfig_format
    for k in ('code-block', 'figure', 'section', 'table'):
        if numfig_format.setdefault(k, k == 'section') is True:
            del numfig_format[k]
    numfig_format.setdefault('num', '%s')
    for k, v in numfig_format.items():
        if v is False:
            if k == 'section':
                raise errors.ConfigError(
                    f"numfig_format: Numbering cannot be disabled for '{k}'")
            numfig_format[k] = no_num


class Num(docutils.ReferenceRole):
    def run(self):
        node = num()
        self.set_source_info(node)
        node['target'] = self.target
        node['title'] = self.title if self.has_explicit_title else '%s'
        if ':' in self.target:
            node['names'].append(self.target)
            self.inliner.document.note_explicit_target(node, node)
        else:
            node['ids'].append(sphinx_nodes.make_id(
                self.env, self.inliner.document, self.target))
        return [node], []


class num(nodes.Inline, nodes.TextElement): pass

# Numbering is global per enumerable node type, and we want per-namespace
# numbering. Numbers are assigned in TocTreeCollector.assign_figure_numbers()
# and immediately compared to the old assignment to invalidate documents,
# and there is no opportunity to modify the numbers between assignment and
# comparison.
#
# We wrap the counters in Cnt instances, which compare equal to all integers, so
# that they aren't taken into account for the invalidation in TocTreeCollector.
# Then, after TocTreeCollector has completed, we update the counter values and
# perform invalidation separately.

class Cnt(int):
    def __eq__(self, other): return isinstance(other, int)
    def __ne__(self, other): return not isinstance(other, int)


class NumCollector(collectors.EnvironmentCollector):
    @staticmethod
    def init(app):
        if not hasattr(app.env, 'tdoc_nums'):
            app.env.tdoc_nums = {}  # docname => id => target
        # Save the old num fignumbers for the invalidation check.
        app.env.tdoc_old_num_fignumbers = {
            doc: nfn for doc, fn in app.env.toc_fignumbers.items()
            if (nfn := fn.get('num')) is not None}

    def clear_doc(self, app, env, docname):
        env.tdoc_nums.pop(docname, None)

    def merge_other(self, app, env, docnames, other):
        for docname in docnames:
            env.tdoc_nums[docname] = other.tdoc_nums[docname]

    def process_doc(self, app, doctree):
        nums = app.env.tdoc_nums[app.env.docname] = {}
        for node in doctree.findall(num):
            nums[node['ids'][0]] = node['target']


def number_per_namespace(app, env):
    # Convert the global numbering to per-namespace numbering.
    per_ns = {}
    for doc, fignums in env.toc_fignumbers.items():
        if (fignums := fignums.get('num')) is None: continue
        nums = env.tdoc_nums[doc]
        for nid, n in fignums.items():
            ns = nums[nid].split(':', 1)[0]
            *sect, cnt = n
            per_ns.setdefault(ns, {}).setdefault(tuple(sect), []) \
                .append((cnt, doc, nid))
    for ns, sects in per_ns.items():
        for sect, cnt_nids in sects.items():
            cnt_nids.sort()
            for i, (cnt, doc, nid) in enumerate(cnt_nids, 1):
                env.toc_fignumbers[doc]['num'][nid] = (*sect, Cnt(i))

    # Invalidate docs for which num fignumbers have changed. The section part
    # has already been checked by TocTreeCollector, so we only need to compare
    # the counter part.
    old = env.tdoc_old_num_fignumbers
    del env.tdoc_old_num_fignumbers
    rewrite = []
    for doc, fignums in env.toc_fignumbers.items():
        if not fignums_equal(fignums.get('num'), old.get(doc)):
            rewrite.append(doc)
    return rewrite


def fignums_equal(lhs, rhs):
    if lhs is None or rhs is None: return lhs == rhs
    if len(lhs) != len(rhs): return False
    for nid, ln in lhs.items():
        if (rn := rhs.get(nid)) is None: return False
        if int(ln[-1]) != int(rn[-1]): return False
    return True


def update_num_nodes(app, doctree, docname):
    for node, text in iter_num(app.env, doctree, docname):
        del node['title']
        if node['names']:
            node += [text]
        else:  # Not an explicit target; replace node with its text
            node.parent.replace(node, text)

    # TOCs are extracted on doctree-read, as a transform with priority=880,
    # while toc_fignumbers are only assigned later, in env-get-updated. So the
    # num nodes in the TOCs must be updated separately. Replace the node by its
    # text altogether, to avoid duplicate 'id' attributes.
    for node, text in iter_num(app.env, app.env.tocs[docname], docname):
        node.parent.replace(node, text)


def iter_num(env, doctree, docname):
    fail = not env.config.numfig
    for node in doctree.findall(num):
        if fail: raise errors.ConfigError("{num}: numfig is disabled")
        try:
            n = env.toc_fignumbers[docname]['num'][node['ids'][0]]
        except KeyError:
            _log.warning("The {num} role cannot be used in orphaned documents",
                         location=node)
            continue
        yield node, nodes.Text(node['title'] % '.'.join(map(str, n)))


def visit_num(self, node):
    self.body.append(self.starttag(node, 'span', suffix=''))


def depart_num(self, node):
    self.body.append('</span>')



# Copyright 2025 Remy Blank <remy@c-space.org>
# SPDX-License-Identifier: MIT

import html

from docutils import nodes
from docutils.parsers.rst import directives
from sphinx.util import docutils, logging

from . import __version__, report_exceptions, Role, to_base64

_log = logging.getLogger(__name__)


def setup(app):
    app.add_directive('quiz', Quiz)
    app.add_role('quiz-ph', QuizPh)
    app.add_role('quiz-hint', QuizHint)
    app.add_role('quiz-input', QuizInput)
    app.add_role('quiz-select', QuizSelect)
    app.add_node(quiz, html=(visit_quiz, depart_quiz))
    app.add_node(quiz_ph, html=(visit_quiz_ph, None))
    app.add_node(quiz_hint)
    app.add_node(quiz_input, html=(visit_quiz_input, None))
    app.add_node(quiz_select, html=(visit_quiz_select, None))
    app.connect('html-page-context', add_js)
    return {
        'version': __version__,
        'parallel_read_safe': True,
        'parallel_write_safe': True,
    }


def add_js(app, page, template, context, doctree):
    if doctree and any(doctree.findall(quiz)):
        app.add_js_file('tdoc/quiz.js', type='module')


class quiz_input(nodes.Inline, nodes.Element): pass
class quiz_select(nodes.Inline, nodes.Element): pass

field_types = (quiz_input, quiz_select)


class Quiz(docutils.SphinxDirective):
    optional_arguments = 2
    option_spec = {
        'class': directives.class_option,
        'style': directives.unchanged,
    }
    has_content = True

    @report_exceptions
    def run(self):
        typ = self.arguments[0] if len(self.arguments) > 0 else 'static'
        gen = self.arguments[1] if len(self.arguments) > 1 else None
        if typ == 'static':
            pass
        elif typ == 'table':
            if gen is None:
                raise Exception("{quizz} table: Missing generator argument")
        else:
            raise Exception(f"{{quizz}}: Invalid type: {typ}")

        children = self.parse_content_to_nodes()
        if any(True for c in children for n in c.findall(quiz)):
            raise Exception("{quiz}: Must not contain {quiz}")
        if not any(True for c in children
                   for n in c.findall(lambda n: isinstance(n, field_types))):
            raise Exception("{quiz}: Must contain at least one field")
        if typ == 'table':
            if not sum(1 for c in children if isinstance(c, nodes.table)):
                raise Exception("{quizz} table: Must contain a single table")
            names = set()
            for c in children:
                for n in c.findall(
                        lambda n: isinstance(n, field_types + (quiz_ph,))):
                    if (name := n['text']) in names:
                        raise Exception("{{quizz}}: Duplicate placeholder or "
                                        f" field name: {name}")
                    names.add(name)

        # Associate hints with fields.
        for child in children:
            for field in child.findall(lambda n: isinstance(n, field_types)):
                for n in field.findall(include_self=False, descend=False,
                                       siblings=True):
                    if isinstance(n, nodes.Text) and not n.strip(): continue
                    if not isinstance(n, quiz_hint): break
                    field['hint'] = n['text']
                    n.parent.remove(n)
                    break
        for child in children:
            for n in child.findall(quiz_hint):
                _log.error("{quiz-hint}: must immediately follow a field",
                           location=n)
                n.parent.remove(n)

        node = quiz('', *children)
        self.set_source_info(node)
        node['type'] = typ
        if gen is not None: node['gen'] = gen
        node['classes'] += self.options.get('class', [])
        if v := self.options.get('style', '').strip(): node['style'] = v
        return [node]


class quiz(nodes.Body, nodes.Element): pass


def visit_quiz(self, node):
    attrs = {'data-type': node['type']}
    if v := node.get('gen'): attrs['data-gen'] = v
    if v := node.get('style'): attrs['style'] = v
    self.body.append(self.starttag(
        node, 'div', suffix='', classes=['tdoc-quiz'], **attrs))
    self.body.append(
        '<div class="content"><span class="tdoc-quiz-hint"></span>')


def depart_quiz(self, node):
    self.body.append('</div>')
    if node.get('type') == 'static':
        self.body.append("""\
<div class="controls">\
<button class="tdoc-check fa-check" title="Check answers"></button>\
</div>\
""")
    self.body.append('</div>')


class QuizPh(Role):
    def run(self):
        node = quiz_ph()
        self.set_source_info(node)
        node['text'] = self.text
        return [node], []


class quiz_ph(nodes.Inline, nodes.Element): pass


def visit_quiz_ph(self, node):
    self.body.append(self.starttag(
        node, 'span', suffix='', classes=['tdoc-quiz-ph'],
        **{'data-text': node['text']}))
    raise nodes.SkipNode


class QuizHint(Role):
    def run(self):
        node = quiz_hint()
        self.set_source_info(node)
        node['text'] = self.text
        return [node], []


class quiz_hint(nodes.Inline, nodes.Element): pass


class QuizField(Role):
    options = {
        'check': directives.unchanged,
        'right': directives.unchanged,
        'style': directives.unchanged,
    }
    content = True

    @report_exceptions
    def run(self):
        node = self.node_type()
        self.set_source_info(node)
        node['role'] = self.name
        node['text'] = self.text
        node['classes'] = self.options['classes'][:]
        if v := self.options.get('check'): node['check'] = v
        style = []
        if (v := self.options.get('right')) is not None:
            node['classes'].append('right')
            if v := v.strip():
                if not v.endswith(';'): v += ';'
                style.append(v)
        if v := self.options.get('style', '').strip():
            if not v.endswith(';'): v += ';'
            style.append(v)
        if style: node['style'] = ' '.join(style)
        self.update_node(node)
        return [node], []

    def update_node(self, node): pass


def attributes(node):
    attrs = {'data-role': node['role'], 'data-text': to_base64(node['text'])}
    if v := node.get('style'): attrs['style'] = v
    if v := node.get('check'): attrs['data-check'] = v
    if v := node.get('hint'): attrs['data-hint'] = v
    return attrs


class QuizInput(QuizField):
    node_type = quiz_input


def visit_quiz_input(self, node):
    self.body.append(self.starttag(
        node, 'input', suffix='', type='text', classes=['tdoc-quiz-field'],
        autocapitalize="off", autocomplete="off", autocorrect="off",
        spellcheck="false", **attributes(node)))
    raise nodes.SkipNode


class QuizSelect(QuizField):
    node_type = quiz_select
    options = {
        **QuizField.options,
        'options': directives.unchanged,
    }

    def update_node(self, node):
        if (opts := self.options.get('options')) is None:
            raise Exception("{quiz-select}: No :options: specified")
        opts = node['options'] = [''] + opts.split('\n')


def visit_quiz_select(self, node):
    self.body.append(self.starttag(
        node, 'select', suffix='', classes=['tdoc-quiz-field'],
        **attributes(node)))
    for opt in node['options']:
        self.body.append(
            f'<option value="{self.attval(opt)}">{html.escape(opt)}</option>')
    self.body.append('</select>')
    raise nodes.SkipNode



#!/usr/bin/env python
# Copyright 2025 Remy Blank <remy@c-space.org>
# SPDX-License-Identifier: MIT

from concurrent import futures
import io
import os
import pathlib
import re
import shlex
import shutil
import stat
import subprocess
import sys
import threading

github_org = 'https://github.com/t-doc-org'


class Error(Exception): pass


def main(argv, stdin, stdout, stderr):
    base = pathlib.Path(argv[0]).parent.resolve().parent
    os.chdir(base)
    tests = base / 'tmp' / 'tests'

    # Find repository names and URLs.
    names = argv[1:]
    if not names:
        names = [d.name for d in base.parent.iterdir()
                 if (d / 'run.py').exists()
                    and (d / 'docs' / 'conf.py').exists()]
    repos = {n: f'{github_org}/{n}' for n in names}
    width = max(len(repo) for repo in repos)
    def label(repo): return f"{repo:{width}} | "

    lock = threading.Lock()
    def write(text):
        with lock: stdout.write(text)

    def on_error(fn, path, e):
        if path == tests: return
        try:
            # Git on Windows sets some files read-only. Remove the flag and
            # try again.
            os.chmod(path, stat.S_IREAD | stat.S_IWRITE, follow_symlinks=False)
            fn(path)
        except OSError:
            stderr.write(f"ERROR: {fn.__name__}: {path}: {e}\n")
    shutil.rmtree(tests, onexc=on_error)
    tests.mkdir(parents=True, exist_ok=True)
    try:
        write("Building wheel\n")
        wheel = build_wheel(tests)

        # Run tests.
        port = 8001
        tasks = {}
        with futures.ThreadPoolExecutor(max_workers=len(repos)) as ex:
            for repo, url in sorted(repos.items()):
                def test(repo=repo, port=port):
                    run_tests(tests, repo, label(repo), url, port, wheel, write)
                tasks[repo] = ex.submit(test)
                port += 1

        # Display output of failures.
        for repo in sorted(repos):
            t = tasks[repo]
            if (e := t.exception()) is None: continue
            se = str(e)
            eol = "" if se.endswith("\n") else "\n"
            write(f"\n{'=' * 79}\n{label(repo)}FAIL\n{'=' * 79}\n{se}{eol}")

        # Display summary.
        write("\n")
        for repo in sorted(repos):
            t = tasks[repo]
            e = t.exception()
            write(f"{label(repo)}{'PASS' if e is None else 'FAIL'}\n")
    finally:
        shutil.rmtree(tests, onexc=on_error)
    return 1 if any(t.exception() is not None for t in tasks.values()) else 0


wheel_re = re.compile(r't_doc_common-[^ ]+\.whl')

def build_wheel(tests):
    out = run(sys.executable, '-P', '-m', 'build', '--no-isolation',
              '--outdir', tests)
    if (m := wheel_re.search(out)) is None:
        raise Exception("Failed to determine wheel name")
    wheel = tests / m.group(0)
    if not wheel.is_file(): raise Exception("Wheel not found")
    return wheel


serving_re = re.compile(r'(?m)^Serving at <([^>]+)>$')

def run_tests(tests, repo, label, url, port, wheel, write):
    # Clone the document repository.
    repo_dir = tests / repo
    write(f"{label}Cloning\n")
    run('git', 'clone', url, repo_dir, '--branch', 'main')

    def vrun(*args, wait=True, **kwargs):
        p = subprocess.Popen(
            (sys.executable, '-P', repo_dir / 'run.py') + args,
            cwd=repo_dir, env={**os.environ, 'TDOC_VERSION': str(wheel)},
            text=True, bufsize=1, stdin=subprocess.DEVNULL,
            stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        if wait:
            out, _ = p.communicate()
            if p.wait() != 0:
                raise Error(f"Command failed: {shlex.join(args)}\n\n{out}")
        return p

    # Get version information. This creates the venv.
    write(f"{label}Getting version information\n")
    vrun('tdoc', 'version', '--debug')

    # Build the HTML.
    write(f"{label}Building HTML\n")
    vrun('tdoc', 'build', '--debug', 'html')

    # Clean the HTML output.
    write(f"{label}Cleaning HTML output\n")
    vrun('tdoc', 'clean', '--debug')

    # Create the store.
    # TODO: Create at version n > 1, then upgrade
    write(f"{label}Creating store\n")
    vrun('tdoc', 'store', 'create', '--debug', '--dev')

    # TODO: Run various commands that access the store

    # Run the local server, wait for it to serve or exit.
    write(f"{label}Running local server\n")
    p = vrun('tdoc', 'serve', '--debug', '--exit-on-failure',
             '--exit-on-idle=2', '--open', f'--port={port}', wait=False)
    try:
        out = io.StringIO()
        for line in p.stdout:
            if out is None:
                write(f"{label}{line}")
                continue
            out.write(line)
            if (m := serving_re.match(line)) is None: continue
            write(f"{label}{line}")
            out = None
    finally:
        write(f"{label}Local server terminated\n")
        if p.wait() == 0: return
        output = f"\n\n{out.getvalue()}" if out is not None else ""
        raise Error(f"The local server has terminated with an error.{output}")


class CommandFailed(Error):
    def __init__(self, e):
        super().__init__(
            f"Command failed with return code {e.returncode}\n"
            f"Command: {shlex.join(str(a) for a in e.cmd)}\n"
            f"\n{e.output}")


def run(*args, **kwargs):
    try:
        return subprocess.run(args, stdin=subprocess.DEVNULL,
                              stdout=subprocess.PIPE, stderr=subprocess.STDOUT,
                              text=True, check=True, **kwargs).stdout
    except subprocess.CalledProcessError as e:
        raise CommandFailed(e)


if __name__ == '__main__':
    try:
        sys.exit(main(sys.argv, sys.stdin, sys.stdout, sys.stderr))
    except SystemExit:
        raise
    except KeyboardInterrupt:
        sys.exit(1)
    except Error as e:
        sys.stderr.write(f'\n{e}\n')
        sys.exit(1)



# Copyright 2025 Remy Blank <remy@c-space.org>
# SPDX-License-Identifier: MIT

jsdelivr = 'https://cdn.jsdelivr.net'

info = {
    'drauu': {
        # https://www.npmjs.com/package/@drauu/core?activeTab=versions
        # https://github.com/antfu/drauu/tags
        'version': '0.4.3',
        'url': lambda v: f'{jsdelivr}/npm/@drauu/core@{v}/dist',
    },
    'mathjax': {
        # https://www.npmjs.com/package/mathjax?activeTab=versions
        # https://github.com/mathjax/MathJax/releases
        'version': '3.2.2',
        'url': lambda v: f'{jsdelivr}/npm/mathjax@{v}/es5',
    },
    'polyscript': {
        # https://www.npmjs.com/package/polyscript?activeTab=versions
        # https://github.com/pyscript/polyscript/tags
        'version': '0.18.2',
        'url': lambda v: f'{jsdelivr}/npm/polyscript@{v}/dist',
    },
    'pyodide': {
        # https://www.npmjs.com/package/pyodide?activeTab=versions
        # https://pyodide.org/en/stable/project/changelog.html
        'version': '0.28.0',
        'url': lambda v: f'{jsdelivr}/pyodide/v{v}/full',
    },
    'sqlite': {
        # https://www.npmjs.com/package/@sqlite.org/sqlite-wasm?activeTab=versions
        # https://github.com/sqlite/sqlite-wasm/releases
        'version': '3.50.1-build1',
        'url': lambda v: f'{jsdelivr}/npm/@sqlite.org/sqlite-wasm@{v}',
    },
}



# Copyright 2024 Remy Blank <remy@c-space.org>
# SPDX-License-Identifier: MIT

import argparse
import datetime
import functools
import itertools
import os
import pathlib
import re
import sys

from . import __project__

_colors = True
try:
    if sys.platform == 'win32':
        import colorama
        colorama.just_fix_windows_console()
except ImportError:
    _colors = False


def want_colors(stdout):
    if not _colors: return False
    if 'NO_COLOR' in os.environ: return False
    if not hasattr(stdout, 'isatty') or not stdout.isatty(): return False
    return True


_ansi_seqs = {
    'NORM': '\x1b[0m',
    'BOLD': '\x1b[1m',

    'BLACK': '\x1b[30m',
    'RED': '\x1b[31m',
    'GREEN': '\x1b[32m',
    'YELLOW': '\x1b[33m',
    'BLUE': '\x1b[34m',
    'MAGENTA': '\x1b[35m',
    'CYAN': '\x1b[36m',
    'WHITE': '\x1b[37m',

    'LBLACK': '\x1b[90m',
    'LRED': '\x1b[91m',
    'LGREEN': '\x1b[92m',
    'LYELLOW': '\x1b[93m',
    'LBLUE': '\x1b[94m',
    'LMAGENTA': '\x1b[95m',
    'LCYAN': '\x1b[96m',
    'LWHITE': '\x1b[97m',
}
_nop_seqs = {k: '' for k in _ansi_seqs}


class AnsiStream:
    """Wrapper around an output stream exposing ANSI sequences as attributes."""
    def __init__(self, stream, color=None):
        self.__stream = stream
        if color is None: color = want_colors(stream)
        self.__tags = _ansi_seqs if color else _nop_seqs

    def __getattr__(self, name):
        try:
            return self.__tags[name]
        except KeyError:
            return getattr(self.__stream, name)


def get_arg_parser(stdin, stdout, stderr):
    """Get an ArgumentParser class operating with the given streams."""
    class Parser(argparse.ArgumentParser):
        def __init__(self, *args,
                     help_description="Show usage information and exit.",
                     **kwargs):
            kwargs['add_help'] = False
            kwargs['allow_abbrev'] = False
            super().__init__(*args, **kwargs)
            self.register('type', 'path', _parse_path)
            self.register('type', 'regexp', _parse_regexp)
            self.register('type', 'timestamp', datetime.datetime.fromisoformat)
            self._positionals.title = "Positional arguments"
            self._optionals.title = "Options"
            if help_description:
                self.add_argument('--help', action='help',
                                  help=help_description)

        def _print_message(self, message, file=None):
            super()._print_message(message, stderr)

        def parse_args(self, *args, **kwargs):
            cfg = super().parse_args(*args, **kwargs)
            cfg.stdin = stdin
            color = None if not hasattr(cfg, 'color') or cfg.color == 'auto' \
                    else cfg.color == 'true'
            cfg.stdout = AnsiStream(stdout, color)
            cfg.stderr = AnsiStream(stderr, color)
            return cfg

    return Parser


def _parse_path(v):
    try:
        return pathlib.Path(v).resolve()
    except OSError as e:
        raise TypeError("Invalid path") from e


def _parse_regexp(v):
    try:
        return re.compile(v)
    except re.error as e:
        raise TypeError("Invalid regexp") from e


_duration_unit_re = re.compile('(us|ms|s|m|h|d|w)')
_duration_kw_map = {
    'us': 'microseconds',
    'ms': 'milliseconds',
    's': 'seconds',
    'm': 'minutes',
    'h': 'hours',
    'd': 'days',
    'w': 'weeks',
}


def parse_duration(duration):
    """Parse a human-readable duration of the form '3d28m32s'."""
    if not duration: raise ValueError(f"Invalid duration: {duration}")
    parts = _duration_unit_re.split(duration)
    if len(parts) % 2 != 0 and not parts[-1]: parts.pop()
    kwargs = {}
    for value, unit in itertools.zip_longest(parts[::2], parts[1::2],
                                             fillvalue='s'):
        k = _duration_kw_map[unit]
        try:
            v = float(value)
        except ValueError:
            raise ValueError(f"Invalid duration: {duration}")
        if v < 0: raise ValueError(f"Invalid duration: {duration}")
        kwargs[k] = kwargs.get(k, 0) + v
    return datetime.timedelta(**kwargs)


def main(fn):
    """Decorate fn as a main() function."""
    @functools.wraps(fn)
    def wrapper(argv=None, stdin=None, stdout=None, stderr=None):
        if argv is None: argv = sys.argv
        if stdin is None: stdin = sys.stdin
        if stdout is None: stdout = sys.stdout
        if stderr is None: stderr = sys.stderr
        try:
            sys.exit(fn(argv, stdin, stdout, stderr))
        except SystemExit:
            raise
        except KeyboardInterrupt:
            sys.exit(1)
        except BaseException as e:
            if '--debug' in argv: raise
            stderr.write(f'\nERROR: {e}\n')
            sys.exit(1)
    return wrapper



# Copyright 2024 Remy Blank <remy@c-space.org>
# SPDX-License-Identifier: MIT

import types

from docutils.parsers.rst import directives
from sphinx.util import docutils, logging

from . import __version__, report_exceptions

_log = logging.getLogger(__name__)


def setup(app):
    app.add_directive('defaults', Defaults)
    return {
        'version': __version__,
        'parallel_read_safe': True,
        'parallel_write_safe': True,
    }


class DefaultsOpts(dict):
    def __bool__(self): return True
    def __contains__(self, k): return True
    def __getitem__(self, k): return lambda v: v


class Defaults(docutils.SphinxDirective):
    required_arguments = 1
    option_spec = DefaultsOpts()
    has_content = False

    @report_exceptions
    def run(self):
        # Find the directive class.
        name = self.arguments[0].lower()
        cls, msgs = directives.directive(name, self.state.memo.language,
                                         self.state.document)
        if cls is None:
            raise Exception(f"{{defaults}}: Unknown directive: {name}")
        if isinstance(cls, (types.FunctionType, types.MethodType)):
            raise Exception(
                f"{{defaults}}: Unsupported function-type directive: {name}")

        # Check option names.
        if unknown := set(self.options) - set(cls.option_spec):
            raise Exception(
                f"{{defaults}}: Unknown options for {{{name}}} directive: "
                f"{", ".join(sorted(unknown))}")

        # Monkey-patch the directive's run() method to set default option
        # values before running the directive.
        if not cls.__dict__.get('_tdoc_defaults_patched'):
            def run(self):
                env = self.state.document.settings.env
                defaults = env.temp_data.get('tdoc_defaults', {}).get(name, {})
                for k, v in defaults.items():
                    self.options.setdefault(k, self.option_spec[k](v))
                return orig_run(self)
            orig_run, cls.run, cls._tdoc_defaults_patched = cls.run, run, True

        # Store the defaults.
        self.env.temp_data.setdefault('tdoc_defaults', {})[name] = self.options
        return []



# Copyright 2025 Remy Blank <remy@c-space.org>
# SPDX-License-Identifier: MIT

import asyncio
from pyodide.ffi import run_sync

from tdoc import core


class Clock:
    """A replacement for pygame.time.Clock."""
    def __init__(self):
        self.t, self.fps_cnt, self.fps = None, 0, 0

    def tick(self, framerate=0):
        if self.t is None:
            t = self.t = self.fps_t = run_sync(core.animation_frame())
            return t
        end = self.t + (1000 / framerate if framerate != 0 else 0)
        while (t := run_sync(core.animation_frame())) < end: pass
        self.t = t
        self.fps_cnt += 1
        if self.fps_cnt >= 10:
            self.fps = self.fps_cnt / ((t - self.fps_t) / 1000)
            self.fps_t, self.fps_cnt = t, 0
        return t

    tick_busy_loop = tick

    def get_time(self): return self.t
    def get_rawtime(self): return core.animation_time()
    def get_fps(self): return self.fps


def get_ticks():
    return core.animation_time()


def wait(milliseconds):
    start = core.animation_time()
    core.sleep(milliseconds / 1000)
    return core.animation_time() - start


def event_wait(timeout=None):
    if timeout is None:
        while (e := pygame.event.poll()) == pygame.NOEVENT: core.sleep(0.02)
        return e
    loop = asyncio.get_running_loop()
    end = loop.time() + timeout / 1000
    while True:
        if (e := pygame.event.poll()) != pygame.NOEVENT: return e
        if (now := loop.time()) >= end: return e
        core.sleep(min(0.02, end - now))


import pygame  # noqa
pygame.time.Clock = pygame.Clock = Clock
pygame.time.get_ticks = get_ticks
pygame.time.wait = pygame.time.delay = wait
# TODO: Test that event_wait() actually works
# pygame.event.wait = event_wait



# Copyright 2024 Remy Blank <remy@c-space.org>
# SPDX-License-Identifier: MIT

import contextlib
import fnmatch
import http.client
import io
import json
import os
import pathlib
import re
import shutil
import subprocess
from urllib import parse
import zipfile

from hatchling.builders.hooks.plugin.interface import BuildHookInterface
from hatchling.metadata.plugin.interface import MetadataHookInterface

LICENSES = 'LICENSES.deps.txt'


class HookMixin:
    @property
    def top(self):
        return pathlib.Path(self.root)

    @property
    def node_modules(self):
        return self.top / 'node_modules'

    @property
    def static_gen(self):
        return self.top / 'tdoc' / 'common' / 'static.gen'

    @property
    def package_lock(self):
        with (self.top / 'package-lock.json').open() as f:
            return json.load(f)


class MetadataHook(MetadataHookInterface, HookMixin):
    def update(self, metadata):
        with (self.top / LICENSES).open('w') as out:
            out.write(f"""\
This file lists the dependencies that are partially or fully included in this
package. The licenses of bundled package archives (e.g. *.whl) can be found in
the archives themselves.
""")
            for path in sorted(self.package_lock['packages']):
                if not path: continue
                root = self.top / path
                try:
                    with (root / 'package.json').open() as f:
                        pkg = json.load(f)
                except OSError:
                    continue
                out.write(f"\n---\n\nName: {pkg['name']}\n"
                          f"Version: {pkg['version']}\n"
                          f"License: {pkg['license']}\n")
                if desc := pkg.get('description'):
                    out.write(f"Description: {desc}\n")
                if repo := pkg.get('repository'):
                    if isinstance(repo, str):
                        out.write(f"Repository: {repo}\n")
                    elif url := repo.get('url'):
                        out.write(f"Repository: {url}\n")
                if url := pkg.get('homepage'):
                    out.write(f"Homepage: {url}\n")
                if author := pkg.get('author'):
                    if isinstance(author, str):
                        out.write(f"Author: {author}\n")
                    else:
                        out.write(f"Author: {author['name']}")
                        if email := author.get('email'):
                            out.write(f" <{email}>")
                        if url := author.get('url'):
                            out.write(f" ({url})")
                        out.write("\n")
                try:
                    license = (root / 'LICENSE').read_text('utf-8').strip()
                except OSError:
                    pass
                else:
                    out.write(f"License text:\n===\n\n{license}\n")


class BuildHook(BuildHookInterface, HookMixin):
    def initialize(self, version, build_data):
        self.app.display_info("Installing node packages")
        npm = shutil.which('npm')
        if npm is None: raise Exception("The 'npm' command cannot be found")
        self.run([npm, 'install'])

        self.app.display_info("Removing generated files")
        shutil.rmtree(self.static_gen, ignore_errors=True)

        self.app.display_info("Generating files")
        os.makedirs(self.static_gen, exist_ok=True)
        self.copy_node('sabayon/dist/sw-listeners.js', 'sabayon-listeners.js')
        self.run([npm, 'run', 'build'])

    def copy_node(self, src, dst):
        shutil.copy2(self.node_modules / src, self.static_gen / dst)

    def copytree_node(self, src, dst, globs=('*',), **kwargs):
        def ignore(path, names):
            return [n for n in names
                    if not any(fnmatch.fnmatch(n, p) for p in globs)]
        shutil.copytree(self.node_modules / src, self.static_gen / dst,
                        symlinks=True, dirs_exist_ok=True, ignore=ignore,
                        **kwargs)

    def run(self, args):
        res = subprocess.run(args, cwd=self.root, stdout=subprocess.PIPE,
                             stderr=subprocess.STDOUT, text=True)
        if res.returncode != 0:
            self.app.abort(f"Command failed:\n{res.stdout}")



# Copyright 2024 Remy Blank <remy@c-space.org>
# SPDX-License-Identifier: MIT

from docutils import nodes
from docutils.parsers.rst import directives
from sphinx.util import docutils, logging

from . import __version__, report_exceptions

_log = logging.getLogger(__name__)


def setup(app):
    app.add_node(iframe, html=(visit_iframe, None))
    app.add_directive('iframe', IFrame)
    app.add_directive('youtube', YouTube)
    return {
        'version': __version__,
        'parallel_read_safe': True,
        'parallel_write_safe': True,
    }


class iframe(nodes.Body, nodes.Element): pass


_attr_names = ['allow', 'referrerpolicy', 'sandbox', 'style', 'title']

def visit_iframe(self, node):
    attrs = {}
    if node['credentialless']: attrs['credentialless'] = ''
    for k in _attr_names:
        if (v := node.get(k)) is not None: attrs[k] = v
    attrs.setdefault('allow', 'autoplay; clipboard-write; encrypted-media; '
                     'fullscreen; picture-in-picture; screen-wake-lock; '
                     'web-share')
    self.body.append(self.starttag(
        node, 'iframe', suffix='', classes=['tdoc'], src=node['src'], **attrs))
    self.body.append('</iframe>\n')
    raise nodes.SkipNode


class IFrame(docutils.SphinxDirective):
    required_arguments = 1
    option_spec = {
        'allow': directives.unchanged,
        'class': directives.class_option,
        'credentialful': directives.flag,
        'referrerpolicy': directives.unchanged,
        'sandbox': directives.unchanged,
        'style': directives.unchanged,
        'title': directives.unchanged,
    }

    @report_exceptions
    def run(self):
        node = iframe()
        self.set_source_info(node)
        node['src'] = self.get_src(self.arguments[0])
        node['classes'] += self.options.get('class', [])
        node['credentialless'] = 'credentialful' not in self.options
        for k in _attr_names:
            if (v := self.options.get(k)) is not None: node[k] = v
        return [node]

    def get_src(self, arg):
        return arg


class YouTube(IFrame):
    def get_src(self, arg):
        if '/' not in arg: return f'https://www.youtube.com/embed/{arg}'
        return arg



#!/usr/bin/env python
# Copyright 2025 Remy Blank <remy@c-space.org>
# SPDX-License-Identifier: MIT

import contextvars
import json
import os
import pathlib
import re
import subprocess
import shutil
import sys
import sysconfig
import threading
import time
import venv

# Set this variable to temporarily install a specific version.
VERSION = ''

package = 't-doc-common'
default_command = ['tdoc', 'serve', '--open']


def main(argv, stdin, stdout, stderr):
    base = pathlib.Path(argv[0]).parent.resolve()

    # Find a matching venv, or create one if there is none.
    version = os.environ.get('TDOC_VERSION', VERSION)
    builder = EnvBuilder(base, version, stderr, '--debug' in argv)
    envs, matching = builder.find()
    if not matching:
        stderr.write("Installing...\n")
        env = builder.new()
        env.create()
        stderr.write("\n")
    else:
        env = matching[0]

    # Remove old venvs.
    for e in envs:
        if e is not env: e.remove()

    # Check for upgrades, and upgrade if requested.
    cur, new = env.upgrade
    if new is not None:
        stderr.write(f"""\
An upgrade is available: {package} {cur} => {new}
Release notes: <https://common.t-doc.org/release-notes.html\
#release-{new.replace('.', '-')}>
""")
        if VERSION:
            stderr.write("Unset VERSION in run.py and restart the program to "
                         "upgrade.\n\n")
        elif version:
            stderr.write("Unset TDOC_VERSION and restart the program to "
                         "upgrade.\n\n")
        else:
            stderr.write("Would you like to install the upgrade (y/n)? ")
            stderr.flush()
            resp = input().lower()
            stderr.write("\n")
            if resp in ('y', 'yes', 'o', 'oui', 'j', 'ja'):
                stderr.write("Upgrading...\n")
                new_env = builder.new()
                try:
                    new_env.create()
                    env = new_env
                except Exception:
                    stderr.write("\nThe upgrade failed. Continuing with the "
                                 "current version.\n")
                stderr.write("\n")

    # Start the upgrade checker.
    threading.Thread(target=env.check_upgrade, daemon=True).start()

    # Run the command.
    args = argv[1:] if len(argv) > 1 else default_command
    args[0] = env.bin_path(args[0])
    return subprocess.run(args).returncode


class lazy:
    def __init__(self, fn):
        self.fn = fn

    def __get__(self, instance, owner=None):
        if instance is None: return self
        res = self.fn(instance)
        setattr(instance, self.fn.__name__, res)
        return res


class Namespace(dict):
    def __getattr__(self, name):
        return self[name]


class Env:
    requirements_txt = 'requirements.txt'
    upgrade_txt = 'upgrade.txt'

    env = contextvars.ContextVar('env')

    def __init__(self, path, builder):
        self.path, self.builder = path, builder

    @lazy
    def requirements(self):
        try:
            return (self.path / self.requirements_txt).read_text()
        except OSError:
            return None

    @lazy
    def sysinfo(self):
        vars = {'base': self.path, 'platbase': self.path,
                'installed_base': self.path, 'intsalled_platbase': self.path}
        return (pathlib.Path(sysconfig.get_path('scripts', scheme='venv',
                                                vars=vars)),
                sysconfig.get_config_vars().get('EXE', ''))

    def bin_path(self, name):
        scripts, ext = self.sysinfo
        return scripts / f'{name}{ext}'

    def create(self):
        self.builder.root.mkdir(exist_ok=True)
        token = self.env.set(self)
        try:
            self.builder.create(self.path)
        except BaseException:
            self.remove()
            raise
        finally:
            self.env.reset(token)

    def remove(self):
        try:
            if self.path.relative_to(self.builder.root) == pathlib.Path():
                raise Exception(f"{self.path}: Not below venv root")
            def on_error(fn, path, e):
                self.builder.out.write(f"ERROR: {fn}: {path}: {e}\n")
            shutil.rmtree(self.path, onexc=on_error)
        except Exception as e:
            self.builder.out.write(f"ERROR: {e}")

    def pip(self, *args, json_output=False, **kwargs):
        p = subprocess.run((self.bin_path('python'), '-P', '-m', 'pip',
                            '--require-virtualenv') + args,
                           stdin=subprocess.DEVNULL, **kwargs)
        if p.returncode != 0: raise Exception(p.stderr)
        if not json_output: return p.stdout
        return json.loads(p.stdout, object_pairs_hook=Namespace)

    @lazy
    def upgrade(self):
        try:
            upgrade = (self.path / self.upgrade_txt).read_text('utf-8')
            return upgrade.split(' ')[:2]
        except Exception:
            return None, None

    def check_upgrade(self):
        try:
            pkgs = self.pip('list', '--format=json',
                            capture_output=True, text=True, json_output=True)
            for pkg in pkgs:
                if pkg.name == package:
                    if pkg.get('editable_project_location') is not None: return
                    cur = pkg.version
                    break
            else:
                return
            data = self.pip('install', '--dry-run', '--quiet', '--upgrade',
                            '--upgrade-strategy=only-if-needed',
                            '--only-binary=:all:', '--report=-', package,
                            capture_output=True, text=True, json_output=True)
            upgrades = {pkg.metadata.name: pkg.metadata.version
                        for pkg in data.install}
            if (new := upgrades.get(package)) is None: return
            (self.path / self.upgrade_txt).write_text(f'{cur} {new}', 'utf-8')
            self.upgrade = cur, new
        except Exception:
            if self.builder.debug: raise


class EnvBuilder(venv.EnvBuilder):
    venv_root = '_venv'
    venv_dev = 'dev'
    venv_prefix = 'venv'

    def __init__(self, base, version, out, debug):
        super().__init__(with_pip=True)
        self.root = base / self.venv_root
        self.version, self.out, self.debug = version, out, debug
        self.requirements = (
            f'-e {base.as_uri()}' if version == 'dev'
            else version if version.startswith(('https:', 'file:'))
            else p.as_uri() if (p := pathlib.Path(version).resolve()).is_file()
            else f'{package}=={version}' if version else package)

    def find(self):
        pat = self.venv_dev if self.version == 'dev' \
              else f'{self.venv_prefix}-*'
        envs = [Env(path, self) for path in self.root.glob(pat)]
        envs.sort(key=lambda e: e.path, reverse=True)
        matching = [e for e in envs if e.requirements == self.requirements]
        return envs, matching

    def new(self):
        name = 'dev' if self.version == 'dev' \
               else f'{self.venv_prefix}-{time.time_ns():024x}'
        return Env(self.root / name, self)

    def post_setup(self, ctx):
        super().post_setup(ctx)
        env = Env.env.get()
        rpath = env.path / f'{env.requirements_txt}.tmp'
        rpath.write_text(self.requirements)
        env.pip('install', '--only-binary=:all:', '--requirement', rpath,
                check=True, stdout=self.out, stderr=self.out)
        rpath.rename(env.path / env.requirements_txt)


MAX_WPATH = 32768
PROCESS_QUERY_INFORMATION = 0x0400
PROCESS_VM_READ = 0x0010

own_process_re = re.compile(r'(?i).*\\(?:py|python[0-9.]*)\.exe$')


def maybe_wait_on_exit(stderr):
    if sys.platform != 'win32': return
    import ctypes.wintypes
    kernel32 = ctypes.WinDLL('kernel32', use_last_error=True)
    DWORD = ctypes.wintypes.DWORD

    # Check if there are any other console processes besides our own
    # (python.exe and potentially py.exe).
    pids = (DWORD * 2)()
    count = kernel32.GetConsoleProcessList(pids, len(pids))
    if count > 2: return

    # There is at least one process besides python.exe. Check if it's py.exe or
    # the shell.
    psapi = ctypes.WinDLL('psapi', use_last_error=True)
    path = ctypes.create_unicode_buffer(MAX_WPATH)
    for pid in pids:
        h = kernel32.OpenProcess(
            DWORD(PROCESS_QUERY_INFORMATION | PROCESS_VM_READ),
            False, pid)
        if h == 0: continue
        try:
            size = psapi.GetModuleFileNameExW(h, None, path, MAX_WPATH)
            if size == 0: continue
            if own_process_re.match(path[:size]) is not None: count -= 1
        finally:
            kernel32.CloseHandle(h)
    if count > 0: return
    stderr.write("\nPress ENTER to exit.")
    stderr.flush()
    input()


if __name__ == '__main__':
    try:
        sys.exit(main(sys.argv, sys.stdin, sys.stdout, sys.stderr))
    except SystemExit:
        raise
    except KeyboardInterrupt:
        sys.exit(1)
    except BaseException as e:
        if '--debug' in sys.argv: raise
        sys.stderr.write(f'\nERROR: {e}\n')
        maybe_wait_on_exit(sys.stderr)
        sys.exit(1)



# Copyright 2024 Remy Blank <remy@c-space.org>
# SPDX-License-Identifier: MIT

import platform
import sys

from polyscript import xworker
from pyodide_js import version

sys.path.append('/lib/tdoc.zip')
import tdoc.core  # noqa

xworker.sync.ready(f"{platform.python_implementation()}"
                   f" {'.'.join(platform.python_version_tuple())}"
                   f" on {platform.platform()},"
                   f" using Pyodide {version}"
                   f" (polyfill: {xworker.polyfill})")



# Copyright 2024 Remy Blank <remy@c-space.org>
# SPDX-License-Identifier: MIT

import contextlib
import datetime
import functools
import pathlib
import re
import secrets
import sqlite3
import threading
import time

max_id_len = 64  # Maximum length of globally-unique identifiers
max_voters_per_poll = 100  # Maximum number of voters per poll


def to_datetime(nsec):
    if nsec is None: return
    return datetime.datetime.fromtimestamp(nsec / 1e9)


def to_nsec(dt, default=None):
    if dt is None: return default
    return int(dt.timestamp() * 1e9)


def placeholders(args): return ', '.join('?' * len(args))


class Error(Exception): pass
client_errors = (sqlite3.DataError, sqlite3.IntegrityError)


class Seqs(dict):
    __slots__ = ()
    mask = (1 << 32) - 1
    mask2 = mask >> 1

    @classmethod
    def newer_than(cls, a, b):
        return b is None or 0 < ((a - b) & cls.mask) <= cls.mask2

    def merge(self, it):
        for key, seq in it:
            if self.newer_than(seq, self.get(key)): self[key] = seq


class Connection(sqlite3.Connection):
    def __enter__(self):
        self._notify = Seqs()
        return super().__enter__()

    def __exit__(self, typ, value, tb):
        res = super().__exit__(typ, value, tb)
        if typ is None and self._notify:
            self._store.notify(self._notify)
        del self._notify
        return res

    def notify(self, *keys):
        for key in keys:
            self._notify[key], = self.row("""
                insert into notifications (key, seq) values (?, 1)
                on conflict do update set seq = (seq + 1) & ?
                returning seq
            """, (key, Seqs.mask))

    def notifications(self, keys):
        return Seqs(self.execute(f"""
            select key, seq from notifications
            where key in ({placeholders(keys)})
        """, keys))

    def row(self, sql, params=(), default=None):
        for row in self.execute(sql, params): return row
        return default

    def meta(self, key, default=None):
        for value, in self.execute(
                "select value from meta where key = ?", (key,)):
            return value
        return default

    @property
    def dev(self):
        return bool(self.meta('dev', False))

    def check_origin(self, origin):
        if not origin and not self.dev:
            raise Exception("No origin specified")

    @functools.cached_property
    def users(self): return Users(self)

    @functools.cached_property
    def tokens(self): return Tokens(self)

    @functools.cached_property
    def groups(self): return Groups(self)

    @functools.cached_property
    def solutions(self): return Solutions(self)

    @functools.cached_property
    def polls(self): return Polls(self)


class DbNamespace:
    def __init__(self, db): self.db = db
    def __getattr__(self, name): return getattr(self.db, name)


class Users(DbNamespace):
    def member_of(self, origin, user, group):
        if user is None: return False
        return bool(self.row("""
            select exists(
                select 1 from user_memberships
                where (origin, user) = (?, ?) and (group_ = ? or group_ = '*')
            )
        """, (origin, user, group))[0])

    def info(self, origin, user):
        name, = self.row("select name from users where id = ?", (user,))
        groups = [g for g, in self.execute("""
            select group_ from user_memberships where (origin, user) = (?, ?)
        """, (origin, user))]
        return {'name': name, 'groups': groups}

    def list(self, users=''):
        return [(name, uid, to_datetime(created))
                for uid, name, created in self.execute("""
                    select id, name, created from users
                    where name regexp ?
                """, (users,))]

    def create(self, names):
        now = time.time_ns()
        uids = [secrets.randbelow(1 << 63) for _ in names]
        self.executemany("""
            insert into users (id, name, created) values (?, ?, ?)
        """, [(uid, name, now) for uid, name in zip(uids, names)])
        return uids

    def memberships(self, origin, users='', transitive=False):
        self.check_origin(origin)
        return list(self.execute("""
            select users.name, group_, transitive from user_memberships
            left join users on users.id = user
            where origin = ? and users.name regexp ?
              and (? or not transitive)
        """, (origin, users, transitive)))


class Tokens(DbNamespace):
    def authenticate(self, token):
        return self.row("""
            select user from user_tokens
            where token = ? and (expires is null or ? < expires)
        """, (token, time.time_ns()), default=(None,))[0]

    def list(self, users='', expired=False):
        now = time.time_ns()
        return [(name, uid, token, to_datetime(created), to_datetime(expires))
                for token, uid, name, created, expires in self.execute("""
                    select token, users.id, users.name, user_tokens.created,
                           expires
                    from user_tokens
                    left join users on users.id = user_tokens.user
                    where users.name regexp ?
                      and (? or expires is null or ? < expires)
                """, (users, expired, now))]

    def create(self, users, expires=None):
        now = time.time_ns()
        expires = to_nsec(expires)
        tokens = [secrets.token_urlsafe() for _ in users]
        self.executemany("""
            insert into user_tokens (token, user, created, expires)
                values (?, (select id from users where name = ?), ?, ?)
        """, [(token, user, now, expires)
              for user, token in zip(users, tokens)])
        return tokens

    def expire(self, tokens, expires=None):
        expires = to_nsec(expires, time.time_ns())
        self.executemany("""
            update user_tokens set expires = ?
            where token = ?
               or exists(select 1 from users where id = user and name = ?)
        """, [(expires, token, token) for token in tokens])


class Groups(DbNamespace):
    def list(self, groups=''):
        return [g for g, in self.execute("""
                    select distinct group_ from user_memberships
                    where group_ regexp :groups
                    union
                    select distinct group_ from group_memberships
                    where group_ regexp :groups
                    union
                    select distinct member from group_memberships
                    where member regexp :groups
                """, {'groups': groups})]

    def members(self, origin, groups='', transitive=False):
        self.check_origin(origin)
        return list(self.execute("""
            select group_, 'user', users.name, transitive from user_memberships
            left join users on users.id = user
            where origin = :origin and group_ regexp :groups
              and (:transitive or not transitive)
            union all
            select group_, 'group', member, false from group_memberships
            where origin = :origin and group_ regexp :groups
        """, {'origin': origin, 'groups': groups, 'transitive': transitive}))

    def memberships(self, origin, groups=''):
        self.check_origin(origin)
        return list(self.execute("""
            select member, group_ from group_memberships
            where origin = ? and member regexp ?
        """, (origin, groups)))

    def modify(self, origin, groups, add_users=None, add_groups=None,
               remove_users=None, remove_groups=None):
        self.check_origin(origin)
        if add_users:
            self.executemany("""
                insert or replace into user_memberships
                    (origin, user, group_, transitive)
                    values (?, (select id from users where name = ?), ?, false)
            """, [(origin, name, group) for name in add_users
                  for group in groups])
        if remove_users:
            self.executemany("""
                delete from user_memberships
                where origin = ?
                  and user in (select id from users where name = ?)
                  and group_ = ?
            """, [(origin, name, group) for name in remove_users
                  for group in groups])
        if add_groups:
            self.executemany("""
                insert or ignore into group_memberships (origin, member, group_)
                    values (?, ?, ?)
            """, [(origin, name, group) for name in add_groups
                  for group in groups])
        if remove_groups:
            self.executemany("""
                delete from group_memberships
                where (origin, member, group_) = (?, ?, ?)
            """, [(origin, name, group) for name in remove_groups
                  for group in groups])
        self.compute_transitive_memberships(origin)

    def compute_transitive_memberships(self, origin):
        items = list(self.execute("""
            select user, group_ from user_memberships
            where origin = ? and not transitive
        """, (origin,)))
        gm = {}
        for member, group in self.execute("""
                    select member, group_ from group_memberships
                    where origin = ?
                """, (origin,)):
            gm.setdefault(member, []).append(group)
        trans = set()
        while items:
            uid, group = it = items.pop()
            if it in trans: continue
            trans.add(it)
            items.extend((uid, g) for g in gm.get(group, []))
        self.execute("""
            delete from user_memberships where origin = ? and transitive
        """, (origin,))
        self.executemany("""
            insert or ignore into
                user_memberships (origin, user, group_, transitive)
                values (?, ?, ?, true)
        """, [(origin, uid, group) for uid, group in trans])


class Solutions(DbNamespace):
    @staticmethod
    def show_key(origin, page):
        return f'solutions:{origin}:{page}'

    def set_show(self, origin, page, show):
        self.execute("""
            insert or replace into solutions (origin, page, show)
                values (?, ?, ?)
        """, (origin, page, show))
        self.notify(self.show_key(origin, page))

    def get_show(self, origin, page):
        return self.row("""
            select show from solutions where (origin, page) = (?, ?)
        """, (origin, page), default=(None,))[0]


class Polls(DbNamespace):
    @staticmethod
    def poll_key(origin, poll):
        return f'polls:{origin}:{poll}'

    @staticmethod
    def voter_key(origin, poll, voter):
        return f'polls:{origin}:{poll}:{voter}'

    def open(self, origin, poll, mode, answers, expires=None):
        self.execute("""
            insert into polls
                (origin, id, mode, answers, expires, results, solutions)
                values (:origin, :poll, :mode, :answers, :expires, 0, 0)
            on conflict do update
                set (mode, answers, expires, results, solutions)
                = (:mode, :answers, :expires, 0, 0)
        """, {'origin': origin, 'poll': poll, 'mode': mode,
             'answers': answers, 'expires': expires})
        self.notify(self.poll_key(origin, poll))

    def close(self, origin, ids):
        self.executemany("""
            insert into polls (origin, id, mode) values (?, ?, null)
            on conflict do update set mode = null
        """, [(origin, p) for p in ids])
        self.notify(*(self.poll_key(origin, p) for p in ids))

    def results(self, origin, ids, value):
        value = bool(value)
        self.executemany("""
            insert into polls (origin, id, results)
                values (:origin, :poll, :results)
            on conflict do update set results = :results
        """, [{'origin': origin, 'poll': p, 'results': value} for p in ids])
        self.notify(*(self.poll_key(origin, p) for p in ids))

    def solutions(self, origin, ids, value):
        value = bool(value)
        self.executemany("""
            insert into polls (origin, id, solutions)
                values (:origin, :poll, :solutions)
            on conflict do update set solutions = :solutions
        """, [{'origin': origin, 'poll': p, 'solutions': value} for p in ids])
        self.notify(*(self.poll_key(origin, p) for p in ids))

    def clear(self, origin, ids):
        phs = placeholders(ids)
        args = (origin, *ids)
        voters = list(self.execute(f"""
            select distinct poll, voter from poll_votes
            where origin = ? and poll in ({phs})
        """, args))
        self.execute(f"""
            delete from poll_votes where origin = ? and poll in ({phs})
        """, args)
        self.notify(*(self.poll_key(origin, p) for p in ids),
                    *(self.voter_key(origin, p, v) for p, v in voters))

    def vote(self, origin, poll, voter, answer, vote):
        if len(voter) > max_id_len: raise Error("Invalid voter ID")
        mode, exp, answers = self.row("""
            select mode, expires, answers from polls
            where (origin, id) = (?, ?)
        """, (origin, poll), default=(None, None, 0))
        if mode is None or (exp is not None and time.time_ns() >= exp):
            raise Error("Voting is closed")
        if answer < 0 or answer >= answers: raise Error("Invalid answer")
        self.notify(self.poll_key(origin, poll),
                    self.voter_key(origin, poll, voter))
        if not vote:
            self.execute("""
                delete from poll_votes
                where (origin, poll, voter, answer) = (?, ?, ?, ?)
            """, (origin, poll, voter, answer))
            return
        if mode != 'multi':
            self.execute("""
                delete from poll_votes
                where (origin, poll, voter) = (?, ?, ?) and answer != ?
            """, (origin, poll, voter, answer))
        self.execute("""
            insert or replace into poll_votes
                (origin, poll, voter, answer) values (?, ?, ?, ?)
        """, (origin, poll, voter, answer))
        voters, = self.row("""
            select count(distinct voter) from poll_votes
            where (origin, poll) = (?, ?)
        """, (origin, poll))
        if voters > max_voters_per_poll: raise Error("Too many voters")

    def poll_data(self, origin, poll, force_results=False):
        mode, exp, results, solutions = self.row("""
            select mode, expires, results, solutions from polls
            where (origin, id) = (?, ?)
        """, (origin, poll), default=(None, None, False, False))
        if exp is not None and time.time_ns() >= exp: mode = None
        voters, votes = self.row("""
            select count(distinct voter), count(*) from poll_votes
            where (origin, poll) = (?, ?)
        """, (origin, poll))
        data = {'open': mode is not None, 'results': bool(results),
                'solutions': bool(solutions), 'voters': voters, 'votes': votes,
                'exp': exp}
        if results or force_results:
            data['answers'] = dict(self.execute("""
                select answer, count(*) from poll_votes
                where (origin, poll) = (?, ?)
                group by answer order by answer
            """, (origin, poll)))
        return data

    def votes_data(self, origin, voter, pids):
        votes = {}
        for p, a in self.execute(f"""
                    select poll, answer from poll_votes
                    where (origin, voter) = (?, ?)
                      and poll in ({placeholders(pids)})
                    order by poll, answer
                """, (origin, voter, *pids)):
            votes.setdefault(p, []).append(a)
        return {'votes': votes}


class ConnectionPool:
    def __init__(self, store, size=None):
        self.store = store
        self.size = size
        self.lock = threading.Lock()
        self.connections = []

    def get(self):
        with self.lock:
            if self.connections: return self.connections.pop()
        return self.store.connect(check_same_thread=False)

    def release(self, db):
        with self.lock:
            if self.size is None or len(self.connections) < self.size:
                db.rollback()
                self.connections.append(db)
            else:
                db.close()


class Waker:
    def __init__(self, cv, limit):
        self.cv = cv
        self.limit = limit
        self._wake = False

    def wait(self, cond, until=None):
        if self.limit is not None and (d := self.limit()) > 0:
            if self.cv.wait_for(cond, d): return
        timeout = None
        if until is not None:
            timeout = until - time.time_ns()
            if timeout <= 0: return
            timeout /= 1_000_000_000
        self.cv.wait_for(lambda: self._wake or cond(), timeout=timeout)
        self._wake = False

    def wake(self):
        with self.cv:
            self._wake = True
            self.cv.notify()


class WakerSet(set):
    __slots__ = ('seq',)

    def __init__(self):
        super().__init__()
        self.seq = None


class Store:
    def __init__(self, path, timeout=10, poll_interval=1):
        self.path = pathlib.Path(path).resolve() if path is not None else None
        self.timeout = timeout
        self.poll_interval = poll_interval
        self.lock = threading.Condition(threading.Lock())
        self.wakers = {}
        self._wake = Seqs()

    def start(self):
        self.dispatcher_db = self.connect(params='mode=ro',
                                          check_same_thread=False)
        if self.path is None: self.create(dev=True)  # Create in-memory DB
        self.dispatcher = threading.Thread(target=self.dispatch)
        with self.lock: self._stop = False
        self.dispatcher.start()

    def stop(self):
        with self.lock:
            self._stop = True
            self.lock.notify()
        self.dispatcher.join()
        self.dispatcher_db.close()

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, typ, value, tb):
        self.stop()

    def connect(self, *, path=False, params='mode=rw', check_same_thread=True):
        if path is False: path = self.path
        uri = f'{path.as_uri()}?{params}' if path is not None \
              else 'file:store?mode=memory&cache=shared'
        # Some pragmas cannot be used or are ineffective within a transaction,
        # and autocommit=False always has a transaction open. Open the
        # connection with autocommit=True, then switch after configuration.
        db = sqlite3.connect(uri, uri=True, timeout=self.timeout,
                             factory=Connection, autocommit=True,
                             check_same_thread=check_same_thread)
        db.execute("pragma journal_mode = wal")
        db.execute("pragma foreign_keys = on")
        db.autocommit = False
        db.create_function(
            'regexp', 2, lambda pat, v: re.search(pat, v) is not None,
            deterministic=True)
        db._store = self
        return db

    def pool(self, **kwargs):
        return ConnectionPool(self, **kwargs)

    @contextlib.contextmanager
    def waker(self, cv, keys, db, limit=None):
        w = Waker(cv, limit)
        with self.lock:
            seq_keys = []
            for key in keys:
                ws = self.wakers.get(key)
                if ws is None:
                    ws = self.wakers[key] = WakerSet()
                    seq_keys.append(key)
                ws.add(w)
            if seq_keys:
                with db: seqs = db.notifications(seq_keys)
                for key, seq in seqs.items():
                    self.wakers[key].seq = seq
        try:
            yield w
        finally:
            with self.lock:
                for key in keys:
                    ws = self.wakers[key]
                    ws.remove(w)
                    if not ws: del self.wakers[key]

    def notify(self, seqs):
        with self.lock:
            self._wake.merge(seqs.items())
            self.lock.notify()

    def dispatch(self):
        next_poll = None
        if self.poll_interval is not None:
            poll_interval = int(self.poll_interval * 1_000_000_000)
            next_poll = time.monotonic_ns() + poll_interval
        db = self.dispatcher_db
        while True:
            with self.lock:
                timeout = (next_poll - time.monotonic_ns()) / 1_000_000_000 \
                          if next_poll is not None else None
                self.lock.wait_for(lambda: self._stop or self._wake,
                                   timeout=timeout)
                if self._stop: break
                wakers = set()
                self._update_waker_seqs(self._wake, wakers)
                self._wake.clear()
                if next_poll is not None and time.monotonic_ns() >= next_poll:
                    if self.wakers:
                        with db: seqs = db.notifications(list(self.wakers))
                        self._update_waker_seqs(seqs, wakers)
                    next_poll = time.monotonic_ns() + poll_interval
            for w in wakers: w.wake()

    def _update_waker_seqs(self, seqs, updated):
        for key, seq in seqs.items():
            if (ws := self.wakers.get(key)) is not None \
                    and Seqs.newer_than(seq, ws.seq):
                ws.seq = seq
                updated.update(ws)

    def backup(self, db, dest):
        if dest.exists():
            raise Exception("Backup destination already exists")
        with contextlib.closing(
                self.connect(path=dest, params='mode=rwc')) as ddb:
            db.backup(ddb)

    def create(self, version=None, dev=False):
        self._check_version(version)
        if self.path is not None and self.path.exists():
            raise Exception("Store database already exists")
        with contextlib.closing(self.connect(params='mode=rwc')) as db:
            return self._upgrade(db, from_version=0, to_version=version,
                                 dev=dev)

    def upgrade(self, db, version=None, on_version=None):
        self._check_version(version)
        from_version = db.meta('version')
        to_version = self._upgrade(db, from_version=from_version,
                                   to_version=version, dev=db.dev,
                                   on_version=on_version)
        return from_version, to_version

    def _check_version(self, version):
        if version is None: return
        for v, _ in self.versions():
            if v == version: return
        raise Exception(f"Invalid store version: {version}")

    def _upgrade(self, db, from_version, to_version, dev, on_version=None):
        now = time.time_ns()
        version = from_version
        for v, fn in self.versions(from_version + 1):
            if to_version is not None and v > to_version: break
            if on_version is not None: on_version(v)
            with db:
                fn(db, dev, now)
                db.execute("""
                    insert or replace into meta (key, value)
                        values ('version', ?)
                """, (v,))
            version = v
        return version

    def versions(self, version=1):
        while True:
            if (fn := getattr(self, f'version_{version}', None)) is None:
                return
            yield version, fn
            version += 1

    def version(self, db):
        version = db.meta('version')
        latest = max(v for v, _ in self.versions())
        return version, latest

    def check_version(self):
        with contextlib.closing(self.connect()) as db, db:
            version, latest = self.version(db)
        if version != latest:
            raise Exception("Store version mismatch "
                            f"(current: {version}, want: {latest})")

    def version_1(self, db, dev, now):
        db.executescript("""
create table meta (
    key text primary key,
    value any
) strict;
create table auth (
    token text primary key,
    perms text not null
) strict;
create table log (
    time int not null,
    location text not null,
    session text,
    data text
) strict;
""")
        db.execute("insert into meta values ('dev', ?)", (bool(dev),))
        if not dev: return
        db.execute("insert into auth (token, perms) values ('*', '*')")

    def version_2(self, db, dev, now):
        db.executescript("""
-- Convert the meta table to "without rowid".
alter table meta rename to _meta;
create table meta (
    key text primary key,
    value any
) strict, without rowid;
insert into meta select * from _meta;
drop table _meta;

-- Create tables for user management.
create table users (
    id integer primary key,
    name text not null unique,
    created integer not null
) strict;
create table user_tokens (
    token text primary key,
    user integer not null,
    created integer not null,
    expires integer,
    foreign key (user) references users (id)
) strict, without rowid;

-- Create tables for group management.
create table user_memberships (
    origin text not null,
    user integer not null,
    group_ text not null,
    transitive integer not null,
    primary key (origin, user, group_),
    foreign key (user) references users (id)
) strict, without rowid;
create table group_memberships (
    origin text not null,
    member text not null,
    group_ text not null,
    primary key (origin, member, group_)
) strict, without rowid;

-- Create table for solutions state management.
create table solutions (
    origin text not null,
    page text not null,
    show text not null,
    primary key (origin, page)
) strict, without rowid;
""")
        if not dev: return
        db.execute("""
insert into users (id, name, created) values (1, 'admin', ?)
""", (now,))
        db.execute("""
insert into user_tokens (token, user, created) values ('admin', 1, ?)
""", (now,))
        db.execute("""
insert into user_memberships (origin, user, group_, transitive)
    values ('', 1, '*', false)
""")

    def version_3(self, db, dev, now):
        db.executescript("""
-- Create tables for poll state.
create table polls (
    origin text not null,
    id text not null,
    mode text,
    answers integer not null default 0,
    expires integer,
    results integer not null default 0,
    solutions integer not null default 0,
    primary key (origin, id)
) strict;
create table poll_votes (
    origin text not null,
    poll text not null,
    voter text not null,
    answer integer not null,
    primary key (origin, poll, voter, answer),
    foreign key (origin, poll) references polls (origin, id)
) strict;
""")

    def version_4(self, db, dev, now):
        db.executescript("""
create table notifications (
    key text primary key,
    seq integer not null
) strict, without rowid;
""")



# Copyright 2024 Remy Blank <remy@c-space.org>
# SPDX-License-Identifier: MIT

import pathlib
import re
import zipfile

from docutils import nodes, statemachine
from docutils.parsers.rst import directives
from sphinx.directives import code
from sphinx.util import display, logging, osutil

from . import __version__, format_attrs, format_data_attrs, names_option, \
    report_exceptions, UniqueChecker

_log = logging.getLogger(__name__)
_base = pathlib.Path(__file__).parent.resolve().parent


def setup(app):
    app.add_directive('exec', Exec)
    app.add_node(exec, html=(visit_exec, depart_exec))
    app.add_env_collector(UniqueChecker('exec-editor',
        lambda doctree: ((n, n.get('editor')) for n in doctree.findall(exec)),
        "{exec}: Duplicate :editor: ID"))
    app.connect('doctree-resolved', check_references)
    app.connect('tdoc-html-page-config', set_html_page_config)
    app.connect('html-page-context', add_js)
    app.add_config_value('tdoc_python_modules', [], 'html', list)
    app.connect('config-inited', set_python_modules)
    app.connect('write-started', write_static_files)
    return {
        'version': __version__,
        'parallel_read_safe': True,
        'parallel_write_safe': True,
    }


class Exec(code.CodeBlock):
    languages = {
        'html': 'html',
        'micropython': 'python',
        'python': 'python',
        'sql': 'sql',
    }

    required_arguments = 1
    optional_arguments = 1
    option_spec = code.CodeBlock.option_spec | {
        'after': names_option,
        'editor': directives.unchanged,
        'include': directives.unchanged_required,
        'output-style': directives.unchanged,
        'style': directives.unchanged,
        'then': names_option,
        'when': lambda c: directives.choice(c, ('click', 'load', 'never')),
    }

    @staticmethod
    def find_nodes(doctree):
        nodes = {}
        for node in doctree.findall(exec):
            nodes.setdefault(node['runner'], []).append(node)
        return nodes

    @report_exceptions
    def run(self):
        if include := self.options.get('include'):
            content = statemachine.StringList()
            for path in include.split():
                rel_path, path = self.env.relfn2path(path)
                self.env.note_dependency(rel_path)
                text = pathlib.Path(path).read_text(self.config.source_encoding)
                content.extend(statemachine.StringList(
                    initlist=text.splitlines(),
                    source=path))
            self.content[:0] = content
        res = super().run()
        for node in res:
            for n in node.findall(nodes.literal_block):
                self._update_node(node)
        return res

    def _update_node(self, node):
        if (hl := self.languages.get(runner := node['language'])) is None:
            raise Exception(f"{{exec}}: Unsupported runner: {runner}")
        node['runner'] = runner
        node['env'] = self.arguments[1] if len(self.arguments) >= 2 else ''
        node['language'] = hl
        node.__class__ = exec
        node.tagname = node.__class__.__name__
        node['classes'] += ['tdoc-exec', f'tdoc-exec-runner-{runner}']
        if v := self.options.get('after'): node['after'] = v
        if v := self.options.get('output-style'): node['output-style'] = v
        if v := self.options.get('style'): node['style'] = v
        if v := self.options.get('then'): node['then'] = v
        node['when'] = self.options.get('when', 'click')
        if (v := self.options.get('editor')) is not None: node['editor'] = v


class exec(nodes.literal_block): pass


def check_references(app, doctree, docname):
    for runner, nodes in Exec.find_nodes(doctree).items():
        names = set()
        for node in nodes:
            names.update(node['names'])
        for node in nodes:
            check_refs(node, names, runner, 'after', doctree)
            check_refs(node, names, runner, 'then', doctree)


def check_refs(node, names, runner, typ, doctree):
    for ref in node.get(typ, ()):
        if ref not in names:
            doctree.reporter.error(
                f"{{exec}} {runner}: Unknown :{typ}: reference: {ref}",
                base_node=node)


def set_html_page_config(app, page, config, doctree):
    if page is None: return
    cfg = {}
    if (md := app.env.metadata[page].get('exec')) is not None:
        cfg['metadata'] = md
    if doctree:
        envs = {}
        for n in doctree.findall(exec):
            if n['when'] == 'never': continue
            envs.setdefault(n['runner'], set()).add(n['env'])
        if envs: cfg['envs'] = {k: sorted(v) for k, v in envs.items()}
    if cfg: config['exec'] = cfg


def add_js(app, page, template, context, doctree):
    if doctree:
        for runner in sorted(Exec.find_nodes(doctree)):
            app.add_js_file(f'tdoc/exec-{runner}.js', type='module')


def set_python_modules(app, config):
    config.tdoc_python_modules.insert(0, str(_base / 'python'))
    if '_python' not in config.tdoc_python_modules \
            and (app.confdir / '_python').exists():
        config.tdoc_python_modules.append('_python')


def write_static_files(app, builder):
    if builder.format != 'html': return

    # Package python modules into a .zip and write it to _static/tdoc.
    with display.progress_message("packaging Python modules..."):
        static = builder.outdir / '_static' / 'tdoc'
        osutil.ensuredir(static)
        zpath = static / 'exec-python.zip'
        zpath.unlink(missing_ok=True)
        with zipfile.ZipFile(zpath, mode='x') as f:
            for mpath in app.config.tdoc_python_modules:
                add_modules(f, app.confdir / mpath)


def add_modules(f, mpath):
    if not mpath.exists(): return
    rel = lambda p: p.relative_to(mpath)
    def on_error(e): raise e
    for root, dirs, files in mpath.walk(on_error=on_error):
        try: dirs.remove('__pycache__')
        except ValueError: pass
        dirs.sort()
        for dn in dirs: f.mkdir(str(rel(root / dn)))
        files.sort()
        for fn in files:
            path = root / fn
            data = path.read_bytes()
            ct = zipfile.ZIP_DEFLATED if data else zipfile.ZIP_STORED
            f.writestr(zipfile.ZipInfo(str(rel(path))), data, compress_type=ct,
                       compresslevel=9)


div_attrs_re = re.compile(r'(?s)^(<div[^>]*)(>.*)$')
pre_attrs_re = re.compile(r'(?s)^(.*<pre[^>]*)(>.*)$')


def visit_exec(self, node):
    try:
        return self.visit_literal_block(node)
    except nodes.SkipNode:
        nameids = node.document.nameids
        after = [nameids[n] for n in node.get('after', ())]
        then = [nameids[n] for n in node.get('then', ())]
        def subst(m): return f'{m.group(1)} {attrs}{m.group(2)}'
        attrs = format_data_attrs(self,
            after=' '.join(after) or None,
            editor=node.get('editor'),
            output_style=node.get('output-style'),
            then=' '.join(then) or None,
            env=node['env'] if node['when'] != 'never' else None,
            when=node['when'])
        if attrs:
            self.body[-1] = div_attrs_re.sub(subst, self.body[-1], 1)
        if attrs := format_attrs(self, style=node.get('style')):
            self.body[-1] = pre_attrs_re.sub(subst, self.body[-1], 1)
        raise


def depart_exec(self, node):
    return self.depart_literal_block(node)



# Copyright 2024 Remy Blank <remy@c-space.org>
# SPDX-License-Identifier: MIT

from http import HTTPStatus
import json
import re

# A regexp matching a hostname component.
hostname_re = r'(?:[a-zA-Z0-9]|[a-zA-Z0-9][a-zA-Z0-9-]{0,61}[a-zA-Z0-9])'


def http_status(status):
    return f'{status} {status.phrase}'


def error(respond, status, msg=None, exc_info=None):
    if msg is None: msg = status.description
    body = msg.encode('utf-8')
    respond(http_status(status), [
        ('Content-Type', 'text/plain; charset=utf-8'),
        ('Content-Length', str(len(body))),
    ], exc_info)
    return [body]


class Error(Exception):
    def __init__(self, status=HTTPStatus.INTERNAL_SERVER_ERROR, msg=None):
        super().__init__(status, msg)

    @property
    def status(self): return self.args[0]

    @property
    def message(self): return self.args[1]


def is_dev(env):
    return env.get('tdoc.dev', False)


def method(env, *allowed):
    m = env['REQUEST_METHOD']
    if allowed and m not in allowed: raise Error(HTTPStatus.METHOD_NOT_ALLOWED)
    return m


def origin(env):
    if (v := env.get('HTTP_ORIGIN')) is None:
        raise Error(HTTPStatus.PRECONDITION_FAILED, "Missing Origin: header")
    return v if not is_dev(env) else ''


def authorization(env):
    if (auth := env.get('HTTP_AUTHORIZATION')) is None: return ''
    parts = auth.split()
    if len(parts) != 2 or parts[0].lower() != 'bearer': return ''
    return parts[1]


def to_json(data, sort_keys=False):
    return json.dumps(data, separators=(',', ':'), sort_keys=sort_keys)


def read_json(env):
    try:
        data = env['wsgi.input'].read(int(env.get('CONTENT_LENGTH', -1)))
        return json.loads(data)
    except Exception as e:
        raise Error(HTTPStatus.BAD_REQUEST)


def respond_json(respond, data):
    body = to_json(data).encode('utf-8')
    respond(http_status(HTTPStatus.OK), [
        ('Content-Type', 'application/json'),
        ('Content-Length', str(len(body))),
        ('Cache-Control', 'no-store'),
    ])
    return [body]


def cors(origins=(), methods=(), headers=(), max_age=None):
    if origins == '*':
        def allow_origin(origin): return [('Access-Control-Allow-Origin', '*')]
    elif isinstance(origins, str):
        pat = re.compile(origins)
        def allow_origin(origin):
            return [('Access-Control-Allow-Origin', origin)] \
                   if pat.fullmatch(origin) else []
    else:
        def allow_origin(origin):
            return [('Access-Control-Allow-Origin', origin)] \
                   if origin in origins else []

    achs = []
    if methods:
        achs.append(('Access-Control-Allow-Methods', ', '.join(methods)))
    if headers:
        achs.append(('Access-Control-Allow-Headers', ', '.join(headers)))
    if max_age is not None:
        achs.append(('Access-Control-Max-Age', str(max_age)))

    def decorator(fn):
        def handle(env, respond):
            def respond_with_allow_origin(status, headers, exc_info=None):
                return respond(
                    status, headers + allow_origin(env.get('HTTP_ORIGIN', '')),
                    exc_info)

            if (env['REQUEST_METHOD'] == 'OPTIONS' and
                    env.get('HTTP_ACCESS_CONTROL_REQUEST_METHOD') is not None):
                respond_with_allow_origin(http_status(HTTPStatus.OK), achs)
                return []
            return fn(env, respond_with_allow_origin)
        return handle
    return decorator



# Copyright 2024 Remy Blank <remy@c-space.org>
# SPDX-License-Identifier: MIT

import yaml

from docutils import nodes
from sphinx.util import docutils, logging

from . import __version__, report_exceptions

_log = logging.getLogger(__name__)


def setup(app):
    app.add_node(metadata)
    app.add_directive('metadata', Metadata)
    app.connect('doctree-read', extract_metadata)
    app.connect('html-page-context', add_head_elements)
    return {
        'version': __version__,
        'parallel_read_safe': True,
        'parallel_write_safe': True,
    }


class metadata(nodes.Element): pass


class Metadata(docutils.SphinxDirective):
    has_content = True

    @report_exceptions
    def run(self):
        node = metadata(attrs=yaml.safe_load(
                        ''.join(f'{line}\n' for line in self.content)))
        self.set_source_info(node)
        return [node]


def extract_metadata(app, doctree):
    md = app.env.metadata[app.env.docname]
    nodes = list(doctree.findall(metadata))
    for i, node in enumerate(nodes):
        if i == 0:
            if (attrs := node['attrs']) is not None: md.update(attrs)
        else:
            _log.warning("More than one {metadata} directive in the document",
                         location=node)
        node.parent.remove(node)


def add_head_elements(app, page, template, context, doctree):
    md = app.env.metadata[page]
    add_head_files(app.add_css_file, md.get('styles', []))
    add_head_files(app.add_js_file, md.get('scripts', []))


def add_head_files(add, entries):
    for entry in entries:
        if isinstance(entry, str):
            add(entry, priority=900)
        else:
            kwargs = dict(entry)
            path = kwargs.pop('src', None)
            for k, v in kwargs.items():
                if v is None: kwargs[k] = ''
            add(path, priority=900, **kwargs)



# Copyright 2024 Remy Blank <remy@c-space.org>
# SPDX-License-Identifier: MIT

# Sphinx options.
author = ''
language = 'fr'
smartquotes = False
primary_domain = None
numfig = True
highlight_language = 'text'
nitpicky = True
exclude_patterns = ['_build', '.DS_Store', 'Thumbs.db']

extensions = [
    'myst_parser',
    'sphinx.ext.duration',
    'sphinx.ext.extlinks',
    'sphinx.ext.githubpages',
    'sphinx.ext.graphviz',
    'sphinx.ext.ifconfig',
    'sphinx.ext.imgconverter',
    'sphinx.ext.intersphinx',
    'sphinx.ext.mathjax',
    'sphinx.ext.todo',
    'sphinx_copybutton',
    'sphinx_design',
    'tdoc.common.ext',
    'tdoc.common.ext.defaults',
    'tdoc.common.ext.exec',
    'tdoc.common.ext.iframe',
    'tdoc.common.ext.metadata',
    'tdoc.common.ext.num',
    'tdoc.common.ext.poll',
    'tdoc.common.ext.quiz',
    'tdoc.common.ext.solution',
]

suppress_warnings = [
    'myst.strikethrough',  # Only supported for HTML, but that's all we want
]

# Extension options.
# The graphviz extension embeds SVG images as <object> tags. This causes them
# to be loaded as separate documents, which doesn't play well with the
# Cross-Origin Isolation workaround (the requests don't get intercepted, lack
# the necessary headers, and therefore get blocked). PNG images are embedded
# as <img>, which don't have this issue.
graphviz_output_format = 'png'
todo_include_todos = True

# MyST options.
myst_enable_extensions = {
    'amsmath',
    'attrs_block',
    'attrs_inline',
    'colon_fence',
    'deflist',
    'dollarmath',
    'fieldlist',
    'html_admonition',
    'html_image',
    # 'linkify',
    # 'replacements',
    'strikethrough',
    'substitution',
    'tasklist',
}
myst_dmath_double_inline = True
myst_heading_anchors = 6

# HTML rendering options.
html_show_sphinx = False
html_theme = 't-doc'

# MathJax options.
mathjax3_config = {
    'loader': {'load': ['input/mml']},
    'options': {
        'menuOptions': {'settings': {'inTabOrder': False}},
    },
}

# LaTeX rendering options.
latex_engine = 'xelatex'

# Domain storage options.
tdoc_domain_storage = {
    'origin': 'https://t-doc.org',
}



# Copyright 2024 Remy Blank <remy@c-space.org>
# SPDX-License-Identifier: MIT

import html
from . import core


def esc(v, quote=True):
    if not isinstance(v, str): v = str(v)
    return html.escape(v, quote)


class _AttributesMeta(type):
    def __call__(cls, *args, **kwargs):
        if len(args) > 1:
            raise TypeError(
                f"{cls.__name__} takes at most 1 positional argument")
        if len(args) == 1:
            obj = args[0]
            if isinstance(obj, cls): return obj
            args = ()
            kwargs[next(iter(cls._attrs))] = obj
        return super().__call__(*args, **kwargs)


class _Attributes(metaclass=_AttributesMeta):
    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            if k not in self._attrs:
                raise TypeError(f"Unexpected argument '{k}'")
            setattr(self, k, v)

    def __iter__(self):
        for k, a in self._attrs.items():
            if (v := getattr(self, k, None)) is not None:
                yield f' {a}="{esc(v)}"'


class Stroke(_Attributes):
    _attrs = {
        'color': 'stroke',
        'width': 'stroke-width',
        'opacity': 'stroke-opacity',
        'dash_array': 'stroke-dasharray',
        'dash_offset': 'stroke-dashoffset',
        'line_cap': 'stroke-linecap',
        'line_join': 'stroke-linejoin',
        'miter_limit': 'stroke-miterlimit',
    }
    __slots__ = _attrs.keys()

Stroke.default = Stroke(color='black', width=1, opacity=1)


class Fill(_Attributes):
    _attrs = {
        'color': 'fill',
        'opacity': 'fill-opacity',
        'rule': 'fill-rule',
    }
    __slots__ = _attrs.keys()

Fill.default = Fill(color='black', opacity=1)


class Transform:
    __slots__ = ('renders',)

    def __init__(self):
        self.renders = []

    def matrix(self, a=0, b=0, c=0, d=0, e=0, f=0):
        @self.renders.append
        def render():
            yield f'matrix({a} {b} {c} {d} {e} {f})'
        return self

    def translate(self, x=0, y=0):
        @self.renders.append
        def render():
            yield f'translate({x}'
            if y != 0: yield f' {y}'
            yield ')'
        return self

    def scale(self, x=0, y=None):
        @self.renders.append
        def render():
            yield f'scale({x}'
            if not (y is None or y == x): yield f' {y}'
            yield ')'
        return self

    def rotate(self, angle, x=None, y=None):
        @self.renders.append
        def render():
            yield f'rotate({angle}'
            if x is not None and y is not None: yield f' {x} {y}'
            yield ')'
        return self

    def skew_x(self, angle):
        @self.renders.append
        def render():
            yield f'skewX({angle})'
        return self

    def skew_y(self, angle):
        @self.renders.append
        def render():
            yield f'skewY({angle})'
        return self

    def __iter__(self):
        if not self.renders: return
        yield ' transform="'
        for i, render in enumerate(self.renders):
            if i > 0: yield ' '
            for v in render(): yield esc(v)
        yield '"'


def matrix(*args, **kwargs): return Transform().matrix(*args, **kwargs)
def translate(*args, **kwargs): return Transform().translate(*args, **kwargs)
def scale(*args, **kwargs): return Transform().scale(*args, **kwargs)
def rotate(*args, **kwargs): return Transform().rotate(*args, **kwargs)
def skew_x(*args, **kwargs): return Transform().skew_x(*args, **kwargs)
def skew_y(*args, **kwargs): return Transform().skew_y(*args, **kwargs)


class _Element:
    _slots = ('_id', 'klass', 'style', 'stroke', 'fill')

    def __init__(self, *, klass=None, style=None, stroke=None, fill=None):
        self._id, self.klass, self.style = None, klass, style
        self.stroke, self.fill = Stroke(stroke), Fill(fill)

    @property
    def id(self):
        if self._id is None: self._id = core.new_id()
        return self._id

    @id.setter
    def id(self, value):
        self._id = value

    def _attrs(self):
        if (v := self._id) is not None: yield f' id="{esc(v)}"'
        if (v := self.klass) is not None: yield f' class="{esc(v)}"'
        if (v := self.style) is not None: yield f' style="{esc(v)}"'
        yield from self.stroke
        yield from self.fill


class Shape(_Element):
    _slots = _Element._slots + ('transform',)

    def __init__(self, *, transform=None, **kwargs):
        self.transform = transform
        super().__init__(**kwargs)

    def _attrs(self):
        yield from super()._attrs()
        if (v := self.transform) is not None: yield from v


class Circle(Shape):
    __slots__ = Shape._slots + ('x', 'y', 'r')

    def __init__(self, x, y, r, **kwargs):
        self.x, self.y, self.r = x, y, r
        super().__init__(**kwargs)

    def __iter__(self):
        yield f'<circle cx="{esc(self.x)}" cy="{esc(self.y)}" r="{esc(self.r)}"'
        yield from self._attrs()
        yield '/>'


class Ellipse(Shape):
    __slots__ = Shape._slots + ('x', 'y', 'rx', 'ry')

    def __init__(self, x, y, rx, ry, **kwargs):
        self.x, self.y, self.rx, self.ry = x, y, rx, ry
        super().__init__(**kwargs)

    def __iter__(self):
        yield f'<ellipse cx="{esc(self.x)}" cy="{esc(self.y)}"'
        yield f' rx="{esc(self.rx)}" ry="{esc(self.ry)}"'
        yield from self._attrs()
        yield '/>'


class Line(Shape):
    __slots__ = Shape._slots + ('x1', 'y1', 'x2', 'y2')

    def __init__(self, x1, y1, x2, y2, **kwargs):
        self.x1, self.y1, self.x2, self.y2 = x1, y1, x2, y2
        super().__init__(**kwargs)

    def __iter__(self):
        yield f'<line x1="{esc(self.x1)}" y1="{esc(self.y1)}"'
        yield f' x2="{esc(self.x2)}" y2="{esc(self.y2)}"'
        yield from self._attrs()
        yield '/>'


class Path(Shape):
    __slots__ = Shape._slots + ('path',)

    def __init__(self, *path, **kwargs):
        self.path = path
        super().__init__(**kwargs)

    def __iter__(self):
        yield f'<path d="'
        for i, el in enumerate(self.path):
            if i > 0: yield ' '
            if isinstance(el, (tuple, list)):
                yield esc(','.join(str(it) for it in el))
            else:
                yield esc(el)
        yield '"'
        yield from self._attrs()
        yield '/>'


class _Poly(Shape):
    def __init__(self, *points, **kwargs):
        self.points = points
        super().__init__(**kwargs)

    def __iter__(self):
        yield f'<{self._tag} points="{esc(' '.join(f'{p[0]},{p[1]}'
                                                   for p in self.points))}"'
        yield from self._attrs()
        yield '/>'


class Polygon(_Poly):
    __slots__ = _Poly._slots + ('points',)
    _tag = 'polygon'


class Polyline(_Poly):
    __slots__ = _Poly._slots + ('points',)
    _tag = 'polyline'


class Rect(Shape):
    __slots__ = Shape._slots + ('x', 'y', 'width', 'height', 'rx', 'ry')

    def __init__(self, x, y, width, height, *, rx=None, ry=None, **kwargs):
        self.x, self.y, self.width, self.height = x, y, width, height
        self.rx, self.ry = rx, ry
        super().__init__(**kwargs)

    def __iter__(self):
        yield f'<rect x="{esc(self.x)}" y="{esc(self.y)}"'
        yield f' width="{esc(self.width)}" height="{esc(self.height)}"'
        if (v := self.rx) is not None: yield f' rx="{esc(v)}"'
        if (v := self.ry) is not None: yield f' ry="{esc(v)}"'
        yield from self._attrs()
        yield '/>'


class Text(Shape):
    __slots__ = Shape._slots + ('x', 'y', 'text')

    def __init__(self, x, y, text, **kwargs):
        self.x, self.y, self.text = x, y, text
        super().__init__(**kwargs)

    def __iter__(self):
        yield f'<text x="{esc(self.x)}" y="{esc(self.y)}"'
        yield from self._attrs()
        yield f'>{esc(self.text, False)}</text>'


class Use(Shape):
    __slots__ = Shape._slots + ('href', 'x', 'y')

    def __init__(self, href, *, x=0, y=0, **kwargs):
        if not isinstance(href, str): href = f'#{href.id}'
        self.href, self.x, self.y = href, x, y
        super().__init__(**kwargs)

    def __iter__(self):
        yield f'<use href="{esc(self.href)}" '
        yield f' x="{esc(self.x)}" y="{esc(self.y)}"'
        yield from self._attrs()
        yield '/>'


class Container(Shape):
    _slots = Shape._slots + ('children',)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.children = []

    def add(self, child):
        self.children.append(child)
        return child

    def circle(self, *args, **kwargs):
        return self.add(Circle(*args, **kwargs))

    def ellipse(self, *args, **kwargs):
        return self.add(Ellipse(*args, **kwargs))

    def line(self, *args, **kwargs):
        return self.add(Line(*args, **kwargs))

    def path(self, *args, **kwargs):
        return self.add(Path(*args, **kwargs))

    def polygon(self, *args, **kwargs):
        return self.add(Polygon(*args, **kwargs))

    def polyline(self, *args, **kwargs):
        return self.add(Polyline(*args, **kwargs))

    def rect(self, *args, **kwargs):
        return self.add(Rect(*args, **kwargs))

    def text(self, *args, **kwargs):
        return self.add(Text(*args, **kwargs))

    def group(self, *args, **kwargs):
        return self.add(Group(*args, **kwargs))

    def symbol(self, *args, **kwargs):
        return self.add(Symbol(*args, **kwargs))

    def use(self, *args, **kwargs):
        return self.add(Use(*args, **kwargs))

    def __iter__(self):
        yield f'<{self._tag}'
        yield from self._attrs()
        yield '>'
        for child in self.children: yield from child
        yield f'</{self._tag}>'


class Group(Container):
    __slots__ = Container._slots
    _tag = 'g'


class Symbol(Container):
    __slots__ = Container._slots
    _tag = 'symbol'

    def _attrs(self):
        yield from super()._attrs()
        yield ' overflow="visible"'


class Image(Container):
    __slots__ = Container._slots + ('width', 'height', 'stylesheet')

    def __init__(self, width, height, *, stylesheet=None, **kwargs):
        self.width, self.height, self.stylesheet = width, height, stylesheet
        super().__init__(**kwargs)

    def _attrs(self):
        yield ' xmlns="http://www.w3.org/2000/svg"'
        yield ' xmlns:xlink="http://www.w3.org/1999/xlink"'
        yield f' viewBox="0 0 {esc(self.width)} {esc(self.height)}"'
        yield f' width="{esc(self.width)}" height="{esc(self.height)}"'
        yield from super()._attrs()

    def __iter__(self):
        if stylesheet := self.stylesheet: id = self.id  # Force ID allocation
        yield '<svg'
        yield from self._attrs()
        yield '>'
        if stylesheet: yield f'<style>\n#{id} {{{esc(stylesheet)}}}\n</style>'
        for child in self.children: yield from child
        yield '</svg>'
